#ifndef _diffincl_h_
#define _diffincl_h_
#include"defo.h"
#include"zdynsys.h"
#include"kspolbds.h"

// constants used in the computation of the difference of solutions
// of differential inclusion x'\in f(x) + [-delta,delta]
// and                       y'=f(y)
extern const int Ldelta_lognorm  /*=2 */;  // an approach based on logarithmic norm
extern const int Ldelta_componentwiseestm /*=3*/; // an approach on based on componentwise estimates


// constants used to describe verious approaches to perform one step in the integration
// of differential inclusions
extern const int dfincl_C0simpleLo /*=1*/; // simple C0-method Lohner method
extern const int dfincl_C0thruC1Lo /*=2*/; // deterministic part is computed using C1-Lohner


int ask_inclmethod();
// the choice between the method based on logarithmic norm and conponent-wise estimates

void incl_method_chose(int *ldeltamethod,int *plognorm);
//  if *ldeltamethod==Ldelta_componentwiseestm, then <*plognorm> is ignored

char *incl_onestepmethod(char* name,int maxlen,int method);
// returns the name (a description) of way we perform one step of the integration
// of differential inclusions
//  maxlen - the size of the table name[]

char *inclmethoddescr(char* name,int maxlen,int method);
// returns the name of Ldeta-method - see comments regarding constants
//  Ldelta_lognorm, Ldelta_componentwiseestm
//  maxlen - the size of the table name[]


void selinclonestepmethod(int *onestepmethod,int *orderdf);
// we select from the console the one step method for diff. incl. if necessary we also
//  ask for for the order of C^1-part

class InclOneStepParamType: public OneStepParamType{
// the description of the method used for the onestep of integration of diff. incl.
  public:
    int onestepmethod; // possible values : dfincl_C0simpleLo,dfincl_C0thruC1Lo


  InclOneStepParamType();
  InclOneStepParamType(int pmethod,int porder,int porderdf,
           int pdoC0intersection,int ponestepmethod);
  void virtual write(ostream &report) const;

};

class InclPoincareParType{
  public:
    InclOneStepParamType  onesteppar;
    double tmax;
    double h; // the time step

    InclPoincareParType(double ph,double ptmax,int pmethod,int porder,int porderdf,
           int pdoC0intersection,int ponestepmethod);
    InclPoincareParType();

  void write(ostream &report) const;

};

void copyToParPoincare(InclPoincareParType &source,Par_Poincare &dest);




class SetWithTail{
  public:
    int dim;
    interval *y;
    int tailexists;
    KuraShi_Tail *ptail; // pointer to the tail - this will be in fact
                       // the structure depending on the PDE under consideration
                       // eventually in the future
    int removetail;
    SetWithTail(int pdim,interval py[],int pwithtail,KuraShi_Tail* ptotail);
    SetWithTail(int pdim,int pwithtail,KuraShi_Tail* ptotail);
    //  pwithtail  - 0 - means that we create SetWithTail, with NULL-tail
    //               1 - otherwise, and ptail=ptotail

    SetWithTail(int pdim);
    SetWithTail(const SetWithTail &s);
    SetWithTail();

    ~SetWithTail();

    SetWithTail & operator=(const SetWithTail &source);
    void assigntail(KuraShi_Tail &T);  // it substitues (copies) the current tail by T
                    // and does all neccessary updates

};


void SwT_hull(const SetWithTail &S1,const SetWithTail &S2,SetWithTail &S);
//  On output  S ; such that S_1 \cup S_2 \subset S

int commonpart(const SetWithTail &S1,const SetWithTail &S2,SetWithTail &S,
         int tailintersectiononly);
//  on output 1 is returned if S_1 \cap S_2 \nonempty
//             and S becomes an upper bound for  S1 \cap S2
//       obtaining the exact intersection is imposible for far tail
// otherwise 0 is returned and S is unchanged
//// if tailintersectiononly - we do it for the tails only

int takeintersectionSWT(int xdim,int tailevolution,const interval x[],const interval fx[],
           const SetWithTail &y, const SetWithTail &fy,interval  xnew[],SetWithTail & ynew,
           int *emptyxfx, int *emptyyfy);
//  On output : xnew= x \cap fx, ynew= y \cap fy
//         if(!tailevolution) then ynew=y
//         *emptyxfx=1 iff x \cap fx =\emptyyset
//         *emptyyfy=1 iff y \cap fy = \emptyset
// 1 is returned if the intersection is non-empty
// 0 - otherwise



// x' \in f(x,y(t))
class InclOde: public Ode{
// the main reason for this class being a child of Ode class is :
//  we want to use the Taylor method for it (i.e. all functions)
  public:
    interval y_c[MAX_DIM+1]; //  the value of the center of current prediction
                    // for y[t] - used in all functions computing the vector field
                    //        and its derivatives;   i.e f(x,y_c)
    double y_c_double[MAX_DIM+1];  // = double(y_c)


    int Ldelta_method; //      the method used in LDelta computation
    int Ldelta_norm;  // if the logarithmic norm is used in LDelta
                 // computation, then this parameter specifies the norm

  // inherited form ODE
 //  char infostr[40];  // inherited from ODE
 //  int dim;           // inherited from ODE
 //  int dissipative;
 //  int enc_order;
 // double divsecreg;  // in the section region we divide the current time step
                // by this number


  void virtual incl_field(const interval x[],SetWithTail &yestm,interval f[]);
  interval virtual  incl_vfield(int k,const interval x[],SetWithTail &yestm) = 0;
   //   vfield(k,x)=f_k(x,yestm)
  interval  virtual incl_dvfield(int k, int i,const interval x[],SetWithTail &yestm) = 0;
  //    dvfield(k,i,x,yestm)=\frac{\partial f_k}{\partial x_i} (x,yestm)
  //used in the computation of the difference of the diff. inlcusion and determistic ODE with y=y_c


   // we want here both functions: <field> - computes full vector fiels
  //  and <vfield> - just
  // one coordinate - in case when computing each coordinate separately,
  // will require redoing many times the same auxliary computation


  // the functions below compute f(x,y_c) and its derivatives
  interval virtual  vfield(int k,const interval x[]) = 0;
  double virtual vfield(int k,const double x[]) = 0;
    //   vfield(k,x)=f_k(x)

  interval  virtual dvfield(int k, int i,const interval x[]) = 0;
  double virtual dvfield(int k, int i,const double x[]) = 0;
/// dvfield(k,i,x)=\frac{\partial f_k}{\partial x_i}(x)

  interval virtual   d2vfield(int k, int i, int j,const interval x[]) = 0;
  double virtual  d2vfield(int k, int i, int j,const double x[]) = 0;
   // d2vfield(k,i,j,x)=\frac{\partial^2 f_k}{\partial x_i  \partial x_j}(x)


  int virtual enclosure(const interval d[],interval h0, interval encl[],int test=0)=0;
// if true is returned then:
//    \varphi([0,h0],d) \subset encl ;   0, y=y_c -
//  h0 - can be negative
//
   // test=0 -  when the procedure fails then a message is send
  //             to a screen and files
  // test=1 -  failure does not result in a message to screen




  int virtual incl_enclosure(const interval x[],SetWithTail &yin,interval h0,
     interval enclx[],SetWithTail &encly, int nomessages=0)=0;
// if true is returned then:
//    \varphi([0,h0],(x,yin)) \subset (enclx,encly)
//  on output yencl=inclusion_interval   - the perturbation to the ODE
//  so <inclusion_interval> has to be computed, if necessary.
// For the constant perturbation of the ODE
//   it is just a constant, but for Galerkin projections <inclusion_interval>
// has to be updated in this function

   // if nomessages==0 -  then when the procedure fails then a message is send
  //             to a screen and files
  // if nomessages==1 -  then the failure does not result in message to screen

  int virtual incl_dissOneStep(const interval x0[],interval h,interval xencl[],
                  interval xstep[]);
  // on input: \varphi([0,h],x0) \subset xencl
  // On output in  <xstep> - the bounds for \varphi(h,x)_i
  //           obtained from the dissipativity are returned

  void virtual fromsegdist(interval encl[],interval dist[],interval h)=0;
    // the function assumes that both a set and its image are contained in
    // <encl> for time [0,h]
    //  then computes a distance from segment joinnig end points and a point
    //  from trajectory (with a current y_c)
    //  on return dist[i] = 0 if x_i - coordinate is monotonic inside <encl>
    //            and is nonzero otherwise and equal to the above mentioned
    //            distance between segment and entire trajectory
    // useful in section region

   void  virtual fromsegdist(interval encl[],interval dist[],interval h,
                    interval mac[][MAX_DIM+1])=0;
    //  the function assumes that both a set and its image are contained in
    // <encl> for time [0,h]
    //  then computes distance from segment joining end points and a point
    //  from trajectory, BUT EXPRESSES IT IN NEW COORDINATES m (new coordinates)
    //  - the matrix pcc->intv_m; new coordinates are given by y=m*(x-P)
    //    <encl> - is given on input in old coordinates
    //  on return dist[i] = 0 if x_i - coordinate is monotonic inside <encl>
    //            and is nonzero otherwise and equal to the above mentioned
    //            distance between segment and entire trajectory
    // useful in the section region, when other coordinates are used


  int virtual y_move(interval xencl[],SetWithTail & yencl,interval h,
          SetWithTail & y,SetWithTail &yn)=0;
    //  y - is mapped to yn after time step h; with xencl and yecnl
    // being enclosures for x and y, respectively

  void set_yc(interval new_y_c[]);
  void current_yc(interval y[]);
  void virtual set_ycenter(interval y[]);

  interval virtual delta_vf(int i,const interval x[],SetWithTail & estmy)=0;
  //  computes i-th coordinate of   f_i(x,estmy) - f_i(x,y_c) for x \in <x>
  void virtual delta_field(const interval x[],SetWithTail & estmy,interval deltaf[]);
   //   deltaf= f(x,estmy)-  f(x,y_c)   for x \in <x>

  int virtual incl_compute_Ldelta(interval h, interval w1_encl[],
        interval w2_encl[],
        SetWithTail & wy,interval Ldelta[]);
// computes \Delta (Ldelta) - with symbols w1,w2,wy - as in my notes
//  w1_encl - is a rough enclosure for x1[0,h] - the solution of x'=f(x,y_c)
//  w2_encl - a roght enclosure for x2[0,h] - the solution of x' \in f(x,[y])
//  wy - the rough enclosure for y-variables

  void incl_fromsegdist(interval w1_encl[],interval w2_encl[],
       SetWithTail & wy, interval dist[],interval h);
  //  function computes a distance from segment joinnig end points
  //  of the "unperturbed ode" (i.e.  with fixed y_c)
  // to actually compute where trajectory was during a time step
  // we need:
  //   x - a starting value, \pi_x \varphi(h,x) \subset xs <---- this for
  //                                  the single valued flow
  //  then \pi_x \varphi([0,h],(x,y)) \subset segment(x,xs) + dist,
  // where dist is computed by incl_fromsegdist

  void incl_fromsegdist(interval w1_encl[],interval w2_encl[],
       SetWithTail &wy, interval dist[],interval h, interval m[][MAX_DIM+1]);
   //  function computes a distance from segment joinnig end points
  //  of the "unperturbed ode" (i.e.  with fixed y_c)
  //   BUT EXPRESSES IT IN NEW COORDINATES m (new coordinates)
    //  - the matrix m=pcc->intv_m; new coordinates are given by y=m*(x-P)
    //    w1_encl,w2_encl - are given on input in old coordinates
    // useful in the section region, when other coordinates are used

  void  writeLDeltaMethod(ostream &report);


 private:


  int virtual incl_Ldelta_lognorm(interval h,
        interval w1_encl[],interval w2_encl[],SetWithTail & wy,interval Ldelta[]);
//  computes \Delta (Ldelta) - with symbols w1,w2,wy - as in my notes
//  using an approach based on logarithmic norms
//  w1_encl - is a rough enclosure for x1[0,h] - the solution of x'=f(x,y_c)
//  w2_encl - a roght enclosure for x2[[0,h] - the solution of x' \in f(x,[y])
//  wy - the rough enclosure for y-variables

  int virtual incl_Ldelta_componentwise( interval h,
        interval w1_encl[],interval w2_encl[],SetWithTail & wy,interval Ldelta[]);
//  computes \Delta (Ldelta) - with symbols w1,w2,wy - as in my notes
//  using an approach based on componentwise estimates
//  w1_encl - is a rough enclosure for x1[0,h] - the solution of x'=f(x,y_c)
//  w2_encl - a roght enclosure for x2[[0,h] - the solution of x' \in f(x,[y])
//  wy - the rough enclosure for y-variables

};

// the equation is x' = f(x) + y(t), where y(t) \in inclusion_interval(t)
class  Perturbed_Ode: public InclOde {
  public:
    Perturbed_Ode(Ode *ode,interval pinclusion_interval[],int pLdelta_method,int plognormused);
    Perturbed_Ode(Ode *ode,interval pinclusion_interval[],int pLdelta_method);
    Perturbed_Ode();

    void virtual incl_field(const interval x[],SetWithTail & yestm,interval f[]);
    interval virtual incl_vfield(int k,const interval x[],SetWithTail & yestm);
    //   vfield(k,x)=f_k(x,yestm)
    interval virtual  incl_dvfield(int k, int i,const interval x[],SetWithTail & yestm);
  //    dvfield(k,i,x,yestm)=\frac{\partial f_k}{\partial x_i} (x,yestm)
    //used in the computation of the difference of the diff. inlcusion and determistic ODE with y=y_c


    // we want here both functions: <field> - computes full vector fields
   //  and <vfield> - just
   // one coordinate - in case when computing each coordinate separately,
   // will require redoing many times the same auxliary computation


   // the functions below compute f(x,y_c) and its derivatives
   void virtual field(const double x[],double f[]);
   void virtual field(const interval x[],interval f[]);
  // we want here both functions: <field> - computes full vector fiels
  //  and <vfield> - just
  // one coordinate - in case when computing each coordinate separately,
  // will require redoing many times the same auxliary computation

  interval virtual vfield(int k,const interval x[]);
  double virtual vfield(int k,const  double x[]);
    //   vfield(k,x)=f_k(x)

  interval virtual  dvfield(int k, int i,const interval x[]);
  double virtual dvfield(int k, int i,const double x[]);
  // dvfield(k,i,x)=\frac{\partial f_k}{\partial x_i}(x)

  interval virtual d2vfield(int k, int i, int j,const interval x[]);
  double  virtual d2vfield(int k, int i, int j,const double x[]);
   // d2vfield(k,i,j,x)=\frac{\partial^2 f_k}{\partial x_i  \partial x_j}(x)

  int virtual enclosure(const interval x[],interval h0, interval encl[],int test=0);
// if true is returned then:
//    \varphi([0,h0],x) \subset encl
//  h0 - can be negative
   // test=0 -  when procedure fails then a message is send
  //             to a screen and files
  // test=1 -  failure does not result in message to screen

  void virtual c1enclosure(interval h,const interval xencl[],
                                 interval Vencl[][MAX_DIM+1]);
 // computation of enclosure for variational part
  //  \frac{\partial \varphi}{\partial x} ([0,h],x)=V([0,h],x)
  //  i.e starting from identity
  //might be used in the integration of 'deterministic' part with y=y_c

  int virtual c1dissOneStep(const interval xencl[],interval h,
               interval Vencl[][MAX_DIM+1],interval Vstep[][MAX_DIM+1]);
  // for dissipative ODE's in Vstep returned are  the bounds for
  //         \frac{\partial \varphi_i}{\partial x}_j(h,x0)
  //           obtained from the dissipativity, x0 is the initial condition
  //            for x, and xencl is such that \varphi([0,h],x0) \subset xencl
  //might be used in the integration of 'deterministic' part with y=y_c


   int virtual incl_enclosure(const interval x[],SetWithTail &yin,interval h0,
     interval enclx[],SetWithTail &encly, int nomessages=0);
// if true is returned then:
//    \varphi([0,h0],(x,yin)) \subset (enclx,encly)
//
 // for this end we want  the following condition:
   //    enclx \subset interior(Z)
   //  where: enclx=x + [0,h]*incl_vfield(Z)
//  yencl,yin - don't have the tail, but
//  on output yencl=inclusion_interval   - the perturbation to the ODE
//  also <inclusion_interval> has to be computed, if necessary .
//  In this implementation it is just a constant, but for Galerkin
//  projections <inclusion_interval> has to be upadated
// in this function

   // if nomessages==0 -  then when the procedure fails then a message is send
  //             to a screen and files
  // if nomessages==1 -  then the failure does not result in message to screen

   void virtual fromsegdist(interval encl[],interval dist[],interval h);
    // the function assumes that both a set and its image are contained in
    // <encl> for time [0,h]
    //  then computes distance from segment joinnig end points and a point
    //  from trajectory (with a current y_c) - in this case without any perturbation
    //  on return dist[i] = 0 if x_i - coordinate is monotonic inside <encl>
    //            and is nonzero otherwise and equal to the above mentioned
    //            distance between segment and entire trajectory
    // useful in section region


  void virtual fromsegdist(interval encl[],interval dist[],interval h,
                    interval mac[][MAX_DIM+1]);
    //  the function assumes that both a set and its image are contained in
    // <encl> for time [0,h]
    //  then computes distance from segment joining end points and a point
    //  from trajectory, BUT EXPRESSES IT IN NEW COORDINATES m (new coordinates)
    //  - the matrix pcc->intv_m; new coordinates are given by y=m*(x-P)
    //    <encl> - is given on input in old coordinates
    //  on return dist[i] = 0 if x_i - coordinate is monotonic inside <encl>
    //            and is nonzero otherwise and equal to the above mentioned
    //            distance between segment and entire trajectory
    // useful in the section region, when other coordinates are used



   int virtual y_move(interval /*xencl*/[],SetWithTail &yencl,interval /*h */,
      SetWithTail & /*y*/,SetWithTail & yn);



  interval virtual delta_vf(int i,const interval x[],SetWithTail & /*estmy*/);
  //  computes i-th coordinate of   f_i(x,estmy) - f_i(x,y_c) for x \in <x>
  void virtual delta_field(const interval x[],SetWithTail & /*estmy*/,interval deltaf[]);
   //   deltaf= f(x,estmy)-  f(x,y_c)   for x \in <x>

  protected:
    Ode *pode; // pointer to an ODE
    interval inclusion_interval[MAX_DIM+1]; // the vector field is given by f(x) + y(t)
                       //  this is a global upper bound for y(t) independent
                       //  of t (during the given time step)
                       //  if we want to allow to change |y(t)|  we should
                       //  modify this variable before we call the Taylor method
                       // in the time step - hence  the enclosure
                       // function should set it correctly

    void init_perturbed_ode(Ode *ode,interval pinclusion_interval[],int pLdelta_method,
         int plognormused);


  private:

};


#endif


