#ifndef _func_h_
#define _func_h_
#include"defo.h"
#include"zset.h"
#include"zdynsys.h"




class LinCC;


class Func {
  public:
   int dim;
   double virtual eval(const double []) const  =0  ;
   interval virtual eval(const interval []) const =0 ;
};

class C1Func: public Func{
  public:
    double  virtual eval(const double [])const =0;
    interval virtual eval(const interval []) const=0;
    interval virtual eval(const Set &x) const=0;
    void virtual gradient(const double x[],double grad[]) const=0;
    void virtual gradient(const interval x[],interval grad[]) const=0;
    void virtual intersection(const interval x[],interval x_on_level[]) const;
       // on output   x \\cap \\{z| f(z) ==0 \\} \subset x_on_level
       // used by Poincare maps (really needed in C1 computations)
       // the one here just copies <x> to <x_on_level>
};


class LinearF: public C1Func {
  public:
    //  dim - from  C1Func (from Func)
    double eval(const double x[]) const;  // f(x) = ci[0] + \sum_i=1^dim x[i]*ci[i]
    interval eval(const interval x[]) const;
    interval eval(const Set &x) const;
    void gradient(const double /*x*/[],double grad[]) const;
    void gradient(const interval /*x*/[],interval grad[]) const;
    void virtual intersection(const interval x[],interval x_on_level[]) const;
       // on output   x \\cap \\{z| f(z) ==0 \\} \subset x_on_level
       // used by Poincare maps (really needed in C1 computations)

    LinearF();

    LinearF(int pdim,const double pci[]);
    LinearF(int pdim,const interval pci[]);

  friend  void transformSection(double sec[],const LinCC *lincc);
  friend  void transformSection(interval sec[],const LinCC *lincc);
  friend  void transformSection(LinearF& linfunc,const LinCC *lincc);
    //  the transformation of the section from <x>  to new coordinates <y>
     //  y=M(x-P)



  private:
    void init(int pdim,const double pci[]);
    void init(int pdim,const interval pci[]);

    void doubletointerval();
    void intervaltodouble();


    double ci[MAX_DIM+1];
    interval ici[MAX_DIM+1];
};



void transformSection(double sec[],const LinCC *lincc);
void transformSection(interval sec[],const LinCC *lincc);
void transformSection(LinearF& linfunc,const LinCC *lincc);
    //  the transformation of the section from <x>  to new coordinates <y>
     //  y=M(x-P)



class Coord: public Func{
  public:
    double eval(const double x[]) const ;  // f(x)=x[cno]
    interval eval(const interval x[]) const;
    Coord(int i=1);


  private:
    int cno;
};

class Radial: public Func{
  public:
    double eval(const double x[]) const;  // f(x)=sqrt(x[r1]^2+x[r2]^2)
    interval eval(const interval x[]) const;
    Radial(int i1=1,int i2=1);
  private:
    int r1,r2;
};



// MAPS
class Map{
 public:
  char infostr[40];
  int dim;

  int virtual map( double x[],double fx[])=0;  // on output fx=f(x)
  int virtual map( double x[],double fx[],double df[][MAX_DIM+1])=0;
        // on output fx=f(x), df=df(x)

  int iterate(int iter,const  double x[],double fx[]);
  int iterate(int iter,const  double x[],double fx[],double df[][MAX_DIM+1]);
          // on output fx=f^iter(x), df=df^iter(x)
};


class FMap : public Map{
//  this is a  full map with a version for double and interval computations
  public:
    // int dim; // - is inherited
    // char infostr[40]; // - inherited
    int virtual map( double x[],double fx[])=0;  // on output fx=f(x)


    int virtual map( double x[],double fx[],double df[][MAX_DIM+1])=0;
    int virtual imapDE( interval x[], interval C[][MAX_DIM+1], interval r[],
                                      interval fx[],estimates *estm)=0;
	int virtual imapDE(Set &x,interval fx[],estimates *estm)=0;								  
									  
    int virtual imapDE(interval x[],interval C[][MAX_DIM+1],interval r[],interval fx[],
               interval df[][MAX_DIM+1],estimates *estm)=0;
 	int virtual imapDE(Set &x,interval fx[],interval df[][MAX_DIM+1],estimates *estm)=0;		   



    int virtual imap( interval x[],interval fx[]);  // on output fx=f(x)
    int virtual imapE(interval x[],interval fx[],estimates *estm);
    int virtual imap(interval x[],interval fx[],interval df[][MAX_DIM+1]);
    int virtual imapE(interval x[],interval fx[],interval df[][MAX_DIM+1],estimates *estm);

    int virtual imapD( interval xcenter[], interval C[][MAX_DIM+1],interval r0[],
                 interval fx[]);
       // on output fx=f(xcenter + C*r0)
	
    int virtual imapD(Set &x,interval fx[]);   
	   
    int virtual imapD(interval x[],interval C[][MAX_DIM+1],interval r[],interval fx[],
               interval df[][MAX_DIM+1]);
 	int virtual imapD(Set &x,interval fx[],interval df[][MAX_DIM+1]);		   

//    int iterate(int iter,const  double x[],double fx[]);
//    int iterate(int iter,const double x[],double fx[],double df[][MAX_DIM+1]);
          // on output fx=f^iter(x), df=df^iter(x)

    int iiterate(int iter,const interval x[],interval fx[]);
          // on output fx=f^iter(x)

    int iiterate(int iter,const  interval x[],interval fx[],interval df[][MAX_DIM+1]);
          // on output fx=f^iter(x), df=df^iter(x)
};


class AffineMap: public FMap
// the map is given by x -> A*x +b
{
  public:
   // dim is inherited
    ~AffineMap();
    AffineMap(int pdim,interval A[][MAX_DIM+1],interval b[]);
    AffineMap(int pdim,double A[][MAX_DIM+1],double b[]);
    AffineMap(int pdim,interval A[][MAX_DIM+1]);
    AffineMap(int pdim,double A[][MAX_DIM+1]);

    AffineMap(interval c);
      // one dimensional map c*x

    int virtual map( double x[],double fx[]);  // on output fx=f(x)
    int virtual imap( interval x[],interval fx[]);  // on output fx=f(x)

    int virtual map(double x[],double fx[],double df[][MAX_DIM+1]);
    int virtual imapD(interval x[],interval C[][MAX_DIM+1],interval r[],interval fx[]);
	int virtual imapD(Set &x,interval fx[]);
    int virtual imapD(interval x[],interval C[][MAX_DIM+1],interval r[],interval fx[],
               interval df[][MAX_DIM+1]);
    int virtual imapD(Set &x,interval fx[],interval df[][MAX_DIM+1]);		   
    int virtual imap(interval x[],interval fx[],interval df[][MAX_DIM+1]);

    int virtual imapDE(interval x[],interval C[][MAX_DIM+1],interval r[],
                                      interval fx[],estimates *estm);
	int virtual imapDE(Set &x,interval fx[],estimates *estm);								  
    int virtual imapDE(interval x[],interval C[][MAX_DIM+1],interval r[],interval fx[],
               interval df[][MAX_DIM+1],estimates *estm);
	int virtual imapDE(Set &x,interval fx[],interval df[][MAX_DIM+1],estimates *estm);		   

  private:
    int dataOn;  // tells wether the momory for the arrays has been reserved
    double *d_b;
    interval *intv_b;
    intervalmatrix1 *intv_A;
    doublematrix1 *d_A;
    void inittables(int pdim, interval piA[][MAX_DIM+1],interval pib[],
                 double pdA[][MAX_DIM+1],double pdb[]);
    // the interval version is:  intv_A*x + intv_b
    //  the double version is:   d_A*x + d_b
};


typedef FMap *PFMap;

void mapcomparerignonrig(ostream &report,FMap *PMap,int mapiter,double xb[]);
// compares rigorous and nonrigorous computation for map PMap at xb

#endif /* ifndef _func_h */

