#ifndef _galks_h
#define _galks_h
#include"diffincl.h"
#include"vfkso.h"


class Galerkin_Proj_KuraShi:   public Perturbed_Ode {
  public:
    Galerkin_Proj_KuraShi(KuraShi *kurashi,interval galerr[],int proj_dim,
			  int pLdelta_method, int pLdelta_norm ,intervalmatrix1 *boxes,int par_boxnum,
         KuraShi_Tail *kstail,int puselocgalerr,int ptailevolution);

    Galerkin_Proj_KuraShi(KuraShi *kurashi,int proj_dim,int pLdelta_method, int pLdelta_norm  );
      // the constructor with tail evolution, the dimension of the projection
  // is given by <proj_dim>, kurashi->dim is set to <proj_dim>


    ~Galerkin_Proj_KuraShi();


    int fullcheckstatus(int newstatus);
    // sets the new value for <fullcheckstatus>, returns the previous value
    int fullcheckstatus();
    // returns the current value of full check status


    int get_tailevolution(); // returns the value of <tailevolution>


    int incl_enclosure(const interval x[],SetWithTail &yin,
     interval h0,interval enclx[],SetWithTail &encly, int nomessages/* =0*/);
     // if true is returned then:
   //    \varphi([0,h0],(x,yin)) \subset enclx \times encly
   //   and  inclusion_interval=galerkin errors for this time step
   //
   // if nomessages==0 - then  when procedure fails then a message is send
   //             to a screen and files
   // if nomessages==1 -  the failure does not result in message to screen


    int  y_move(interval xencl[],SetWithTail & yencl,interval h,
          SetWithTail & y,SetWithTail &yn);
    //  y - is mapped to yn after time step h; with xencl and yecnl
    // being enclosures for x and y, respectively


    int  dissOneStep(const interval x0[],interval h,interval xencl[],interval xstep[],
      int useperturbation,const interval perturb_vect[]);
     //  if use_perturbation then we add <perturb_vect> to the vector field of gal. proj.  KS       
     // on input: \varphi([0,h],x0) \subset xencl   
     // on output: in xstep - the bounds for \varphi(h,x)_i
     //           obtained from the dissipativity are returned
     //  these are bounds for Galerkin projection (without inclusion of the tail)

    int  incl_dissOneStep(const interval x0[],interval h,interval xencl[],
            interval xstep[]);
       // on input: \varphi([0,h],x0) \subset xencl   
     // On output: in xstep - the bounds for \varphi(h,x)_i
     //           obtained from the dissipativity are returned
     //  these are bounds for Galerkin projection + influence of the tail



    interval  l2norm(interval x[],SetWithTail &y);
    interval  h1norm(interval x[],SetWithTail &y);
    interval  c0norm(interval x[],SetWithTail &y);
    interval  c1norm(interval x[],SetWithTail &y);
    // above functions return an apriopriate norm of x + galerkin_tail

   

   
   
  protected:
  // inherited from Perturbed_ODE
  //    Ode *pode; // pointer to an ODE
  //    interval inclusion_interval[MAX_DIM+1]; // the vector field is given by f(x) + inclusion_interval
            // here we  the Galerkin errors - the influence of the tail

    int fullenclcheck; // if 0 - then we have really a differential inclusion
              //  1 - we check that in <dim> our trajectory is in boxes
    // or whatever apriori bounds we are using (including an influence of the
               // tail, requires that <fullenclcheck==1>

    int uselocal_galerr;  // if uselocal_galerr - tail does not evolve,
                         // but we recompute
                         // the Galerkin error locally in x

    int tailevolution;  // if tailevolution - we allow for tail to evolve and
                      // the Galerkin errors are computed locally

    int fixedtaildefined; // tells if the 'fixed tail is defined' -
    // the variables below describe it
    // THE  VARIABLES BELOW are used only if !tailevolution
    
    int boxnum; // number of boxes <=100
    intervalmatrix1  *pboxes;


    interval *gl_galerr; // the 'global' Galerkin error - IHull of
                     // galerkin errors over all boxes; this is the 'global'
                     //                                inclusion_interval
    interval *loc_galerr; // the local Galerkin error - it should
                       // be set in 'incl_enclosure' so that - the function
                       // advancing forward diff. inclusion may use it

    KuraShi *pkurashi; // this will be the same pointer as pode,
               // which is inherited from  Perturbed_ODE

    KuraShi_Tail *pks_tail; // this is the pointer to the fixed tail
    // used only if !tailevolution ; 
    // this variable is used only in the the function <compute_loc_galerr>

    void  compute_loc_galerr(interval x[]); // computes Galerkin
                           // errors for x \in x[] and the tail pks_tail
                           // the result is placed in this->loc_galerr

   private:
     interval norm(int nonum,interval x[],KuraShi_Tail *T);
     //  nonum=1,2,3,4 - corresponds respectively to l2,h1,c0 and c1-norms


     int  check_apriori_bounds(const interval x[],interval h0,interval enclx[],
              int test=0);
    // 1 if returned if enclx \subset boxes
      // test=0 -  when procedure fails then a message is send
     //             to a screen and files
     // test=1 -  failure does not result in any error reports


     void  diag_apriori_bounds(const interval x[],interval h0,
         const interval enclx[]);
       // the function writes to the file <errorlogfile>
       // the data from failed enclosure procedure
       // it tries to give a reason for this

     int fixedtail_incl_enclosure(const interval x[],SetWithTail &yin,
				  interval h0,interval enclx[],
                             SetWithTail &encly, int nomessages=0);


     int tevol_incl_enclosure(const interval x[],SetWithTail &yin,
	      interval h0,interval enclx[],SetWithTail &encly, 
              int nomessages=0);
  // the enclosure function with the tail evolution 

};

#endif

