#ifndef _kspolbds_h
#define _kspolbds_h


// 17.VIII.2009 - currently this file contains two very related classes
//  Poly_Bds i KuraShi_Tail - KuraShi_Tail is richer, I need to glue them into one
// class


class Poly_Bds{
// the tail is given by a_k \in a[k] for m < k <= BIG_M
  //                      |a_k| <= C/|k|^s for  k > BIG_M
 public:
    int m, Big_M;
    int s;
    interval c;
    interval *a;

    Poly_Bds(const Poly_Bds &initbds);
    Poly_Bds( int p_m, int p_Big_M,const Poly_Bds &initbds);
    Poly_Bds( int p_m, int p_BIG_M,int p_s, interval p_c,const interval pa[]);
    Poly_Bds( int p_m, int p_BIG_M);
    ~Poly_Bds();

    interval operator()(int i) const;
    //  returns i-th coordinate

    void reset(int p_m,int p_Big_M);

    void newvalues(int p_m, int p_BIG_M,const Poly_Bds &bds);

    void newvalues(int p_m, int p_BIG_M,int p_s, interval p_c,const interval pa[]);

    void newvalues(const Poly_Bds &p); // this=p

    void multiply(interval t);
    // this*=t;

    void save(ostream &report, const char comment[]) const ;
    // saves in a text format

    void save(char* fname, const char comment[]) const;
    // saves in a text format

  private:
    int remove_a; // tells if the destructor has to remove a[] from memory
};

void add_Pol_Bounds(const Poly_Bds &t1,const Poly_Bds &t2, Poly_Bds &r);
// adds bounds:  on output: r=t1+t2

void arr_Pol_Bds(const interval t[][MAX_DIM+1], int tdim,const Poly_Bds &b, Poly_Bds &r);
//  on output r=t*b;   dim(t)=tdim <=b.Big_M



class KuraShi_Tail{
  // the tail is given by a_k \in a[k] for m < k <= BIG_M
  //                      |a_k| <= C/|k|^s for  k > BIG_M
  public:
    int m, big_m;
    int s;
    interval c;
    interval *a;

    KuraShi_Tail( int p_m, int p_BIG_M,int p_s, interval p_c, interval pa[]);
    KuraShi_Tail( int p_m, int p_BIG_M);
    KuraShi_Tail();
    ~KuraShi_Tail();

    KuraShi_Tail(const KuraShi_Tail &source); // constructor
    KuraShi_Tail & operator=(const KuraShi_Tail &source);

    interval operator()(int i) const;
    //  returns i-th coordinate
    void multiply(interval t);
    // this*=t;
    void inflatediam(interval t);
   //  on output - this(i)=center(this(i) old value) + t*(-1,1)*old diameter this->(i)

    void save(ostream &report,int howtowrite, const char comment[]) const;
// if howtowrite == 3, then it is a program form
// otherwise is 'text' form:
//      howtowrite=1 - standard interval format, =2 - mid point + [-d,d]

    void savebin(FILE *fbin) const;
    int readbin(FILE *fbin);

    void newvalues(int p_m, int p_BIG_M,int p_s, interval p_c, interval pa[]);

    int subset(const KuraShi_Tail &T) const;
       // return 1, when this \subset T

    void decpower(int d);
    // we increase the far tail, so new bound is given by C(new)/k^{s-d};
    // hence the new s is equal to s-d

    void newfartail(interval aMp1,int ns);
    // we set a new value for the far tail, such that s(new)=ns
    //  and T(M+1) >= aMp1 (in fact we do it with some margin)
    //   to make sure that some more schemes fo compute aMp1=C/(M+1)^{s}
    //   it will still hold

    void kssymm(); // applies the R-symmetry to the tail variables
             // a_{2k}->a_{2k},  a_{2k+1}-> -a_{2k+1}

  private:
    int remove_a; // tells if the destructor has to remove a[] from memory
};

void KS_Tail_hull(const KuraShi_Tail & T1,const KuraShi_Tail &T2, KuraShi_Tail &T);
//  on output T ; such that T_1 \cup T_2 \subset T


int commonpart(const KuraShi_Tail & T1,const KuraShi_Tail &T2, KuraShi_Tail &T);
//  on output 1 is returned if T_1 \cap T_2 \nonempty
//             and T becomes an upper bound for  T_1 \cap T_2
//       obtaining the exact intersection is imposible for far tail
// otherwise 0 is returned and T is unchanged



#endif  // ndef _kspolbds_h
