#ifndef _perfixp_h
#define _perfixp_h

#include"defo.h"
#include"zdynsys.h"
#include"func.h"
#include"rigpoinc.h"
#include"section.h"

extern const  int meth_newtonsearch;
extern const  int meth_forwarditeration;

char *meth_fixed_point_name(int method,char* name,int maxlen);

void   writePMaptraj(ostream &report,FPoincareMap *PMap,int mapiter,
		     double xb[]);
// writes : Pmap^i(xb) for i=0,,,,mapiter-1
//   xb - is given in the section  coordinates

int odefindfixp(Ode &ode,double x[],double h,int order, double tmax,double err);
// function moves forward <x> up to a time <tmax> in the hope that the
//  solution will converge to a fixed point;
//<h> - the time step
// <order> - the order of the method
//  <err> -is a tolerance parameter
//   point is considered to be a fixed point if meets two criteria
//    |ode.vfield(x)| < err
//     <x> does not move more then <err> in time 0.05*<tmax>
//  if order <=4 then 4-th order Runge-Kutta is used
//               otherwise we use taylor method


int fixpdfanal(Ode &ode,double fixp[],int *no_unstabledir,double *stabfactor);
// <fixp> is a fixed point for <ode>
// On output *no_unstabledir - is the number of unstable directions
//    for linearization at fixp
//      *stabfactor - (for flow) max reale part of eigenvalue of df
// returns 1 - when all went ok
//         0 - otherwise




int newtonfixp(Ode &ode,double x[],double a[MAX_DIM+1][MAX_DIM+1],int maxiterno,
               double err);
  // function looks for zero of the vector field <ode>
  // using Newton-type algorithm
  //  x->N(x) = x - df(x)^{-1}f(x), with df(x)^{-1} replaced by <a>
  // for this algorithm to work <x> - our initial guess must be close to true
  // zero
  //  a - shall be close to df(x)^{-1}
//  ---------
// on output x - contains the fixed point found


int newtonfixp(Ode &ode,double x[],int maxiterno,double err);
  // function looks for zero of the vector field <ode>
  // using Newton-type algorithm
  //  x->N(x) = x - df(x)^{-1}f(x), with df(x)^{-1} replaced by <a>
  // for this algorithm to work <x> - our initial guess must be close to true
  // zero
  //  ---------
// on output x - contains the fixed point found

int newton0_1dim(double (*f)(double),double (*df)(double),double xstart,double *x,
    double precision,int maxiter=100);
// newton algorithm for 1-dim problem f(x)=0
//  x-> x - df^{-1}*f(x)
// xstart - starting value for iteration
// *x - solution of f(x)=0 on output
// precison - if |f(x)| < precison then we consider assume that this x is ok
// maxiter - maximum number of iterates


int newtonsearch(Map &map,int exp,int fixp,double x[],int maxiterno,
    double maxstep,double err,int *iter);
  // function looks for fixed point (or zero) of <map>^<exp>
  // using Newton-type algorithm
  // if <fixp> then we look for fixed point
  //   otherwise we look for zero
  //
  //  x->N(x) = x - df(x)^{-1}f(x)
  // for this algorithm to work <x> - our initial guess must be close
  //  <maxstep> - we don't allow bigger step than this number
  // <err> - when we are this close to fixed point (or zero)(in Euclidean norm),
  //   then we   stop
  //  ---------
// on output x - contains the fixed point (or zero) found
//  *iter - is equal to the number of iterations needed


int findattrfpMap(Map &map,double xs[],int itno,int max_iter, double err,int *iterfound);
// the function searches for an attracting fixed (periodic) point for <map>
//  by a naive iteration , hoping for convergence to fixed point
//   xs - on input - the starting point, on output - the fixed point for
//       <map^itno>
//   max_iter - upper limit for the number of iterates of <map>
//   err - the tolerance (error) in the criterion  for fixed point
//   if the point x is found such that |map(x) - x|_{max} < err, then 1 is returned
//                  xs - contains the fixed point found
//                  *iterfound - the number of iterates
///  if we don't succed then 0 is returned and xs stays unchanged




int improvefixpoint(Map &map,int iterno,double x[],int max_iter,double err);
// On input: x is a candidate for a fixed point for map^iterno
// On output - if everything went ok (1 is returned) x - is better estimate
//           for fixed point of map^{iterno}
////////////
//we first check whether the candidate for the fixed (periodic) point  is attracting -
//        if this is the case then we just iterate forward
// otherwise use the Newton method


#endif  /* _perfixp_h */
