#ifndef _rigkserr_h
#define  _rigkserr_h
#include"coordc.h"
#include"vfkso.h"


class DerData {
  public:
    bool unique_test_on;   // tells if an uniqueness test was performed
    bool unique;  // 1 - when conditions for uniqueness are satisfied
    int uniq_worstco; // coordinate for which we have the worst case in uniqueness
    interval uniq_badmargin; // here is a value how bad it was

    bool lognorm_computed;  // tels whether we computed logarithmic norm
    interval lognorm; // logarithmic norm (in Max - norm)
    int maxlognorm; // the index of the coordinate for which lognorm was maximized
    bool complexeigenvalues;
    interval complex_off_diag;  // the value of the off-diagonal term in complex case
    int dim_tab;
    double alpha; //  alpha >= offdiag/diag from uniqueness computation
    double* tab_l;  // l_i - from logarithmic norm computations
    double* tab_a;  // the values of a_{(i_0)(i)}
    double* tab_u; // the values in the uniqueness criterion
    double* tab_du; // the values of diagonal term in uniqueness computation
    double* tab_dl; // the values of diagonal term in logarithmic norm computation
    double* tab_offdiag; // the values of off-diagonal terms in uniqueness
            // and logarithmic norm computation
    double* tab_offdiag_unstb; // the values of off-diagonal terms in instability
                // computations - this may differ from the off-diag terms
                // in uniqueness and log. norm. comp - because here
                // eventually unstable cells are treated as one cell

    bool instability_test_on; // tells if an instability test was performed
    bool unstable; // tells is if the conditions for instability are satisfied
    double d_unstable; // the value of d_{(i_0)} for an instability test
    int minunstable; // the index for which a_{(i_0)(i)} is the smallest one


    //  begin - related to cone-conditions
    bool cc_test_on; // telles whether we tested cone-conditions
    bool cc_ok;  // tells whether the cone condition is satisfied
    double cc_mingap;  // the worst value of diag(i) - offdiag(i)
    int cc_index_mingap;
    // end - related to cone conditions



    intervalmatrix1 *dfnc;  //df - in new coordinates
    int m_dim; // the m-dimension



    DerData(int pdim_tab);
    ~DerData();

    int setnewdim(int pdim_tab);    // dim_tab=pdim_tab; returns previous value
                                    // deletes old tables and assigns new one
    void assign_tabl(int i, double val); // //  tab_l[i]=val
    void assign_taba(int i, double val); // //  tab_a[i]=val
    void assign_tabu(int i, double val); // //  tab_u[i]=val
    void assign_taboffdiag(int i, double val); // //  tab_offdiag[i]=val
    void assign_tabdu(int i, double val); // //  tab_du[i]=val
    void assign_tabdl(int i, double val); // //  tab_dl[i]=val
    void assign_taboffdiag_unstb(int i, double val); // tab_offdiag_unstb[i]=val


    void wr_derdata(ostream &out,bool fullreport) const;

    #ifdef WIN95
    void wr_derdata(FRAME &out) const;
    #endif
   
};

interval ks_d2B(int k,int i,int j);
// calculates D^2f_k/Dx_iDx_j (x)

interval ks_dB(int i, int j,const interval a[],int Big_M,interval c,int s);
//  returns the partial derivative of the bilinear part of KS
//  dB_i/da_j (a)
//    a - is Big_M-dimensional vector
//     a_k \in a[k] for k=1...Big_M  and |a_k|<= C/|k|^s  for k > Big_M



interval ks_db_nc(int i,int j,const interval a[],int m,int Big_M,interval c,int s,
               const LinCC &coch);
// returns \frac{\partial {\tilde B}_i}{\partial x_j}(a) - of the bilinear term
//  in KS -equations  expressed in the new coordinates
//   a - is Big_M - dimensional vector
//     a_k \in a[k] for k=1...Big_M  and |a_k|<= C/|k|^s  for k > Big_M
//   coch - is the change of coordinates in m-first variables
//      for higher variables the coordinate change is diagonal
//  MUST BE:    (i>m or j>m)
//  the formula is:
//      if   i<=m and j>m   \sum_{k<m} A_{ik} db_k/dx_j
//      if  j>m  and i >m - db_i/da_j
//      if  i>m and  j<=m  - \sum_{l<=m} db_i/da_l * A^{-1}_{lj}
//               where A is 'm'-matrix in coch


void rigboundsinfsum(interval a[],interval C, int s,int m, int big_m,
       interval sum[], interval& D);
// On input: a[i] - a bound for i-th coordinate for i<=big_m
//            for i > big_m the bound is given by: C/i^s*[-1,1]
// On output:  \Sum_{ n \in \mathbb{Z}} a_n a_{k-n} \subset sum[k] for k > m, k<=big_m
//  for k > M (big_m)  |\Sum_{n \in \mathbb{Z}} a_n a_{k-n}| <= D/k^{s-1]
/////////////////////////
/// \Sum_{n \in \mathbb{Z}} a_n a_{k-n} =  FS(k) - 2 * IS(k)
//         where  FS(k)=\sum_{n=1}^{k-1} a_n a_{k-n}
//                IS(K)=\sum_{n=1}^\infty a_n a_{n+k}
/////// the need for this function arise from the following representation
//        of  the KS-equation with odd periodic boundary conditions
//          a_k'=\lambda_k a_k - k*(\Sum_{n \in \mathbb{Z}} a_n a_{k-n})
//      but since a_{-k}=-a_k, then we obtain
//            \Sum_{n \in \mathbb{Z}} a_n a_{k-n} =  FS(k) - 2 * IS(k)



void rigkscompbilinearterm(interval b[],const interval err[],int m, int big_m,interval c,
               int s,interval &d,const interval boxes[][MAX_DIM+1],int boxnum);
  // estimates bilinear term  -k * Sum_{i=-\infty}^{i=\infty} q_i q_{k-i}
  // on OUTPUT b[k] contains bounds for bilinear term for m <k <=big_m
  //                  d/k^{s-2} is a bound for bilinear term for k > big_m
  // including tail - RIGOROUS PROCEDURE
  //  INPUT PARAMETERS:
  // <big_m> - dimension of more carrefull estimates ; err[k]  k <=big_m
  // <m> - which represents the number of first
  //   modes for which bounds are fixed in advance (master variables)
  // <err[k]> - bound for i-th coordinate (when big_m => k> m)
  //  for k > big_m, error estimate is <c>/k^<s>
  // boxes[1..<boxnum>] - is <n> dimensional covering covering of the invariant
  //     set for <m>-dimensional Galerkin projection

void rigkscompbilinearterm(interval b[],const interval err[],int m, int big_m,interval c,
               int s,interval &d,const interval x[]);
  // estimates bilinear term  -k * Sum_{n=-\infty}^{n=\infty} q_n q_{k-n}
  // on OUTPUT b[k] contains bounds for bilinear term for m <k <=big_m
  //                  d/k^{s-2} is a bound for bilinear term for k > big_m
  // including tail - RIGOROUS PROCEDURE
  //  INPUT PARAMETERS:
  // <big_m> -  BIG_M ; err[k]  k <=big_m
  // <m> - which represents the number of main modes
  // the set over which the nonlinear term is computed is given by
  // x - first m- variables
  // <err[k]> - bound for i-th coordinate (when M => k> m)
  //  for k > big_m, error estimate is <c>/k^<s>


void   selfconsbounds(const interval inerr[],interval err[],interval c1,interval &c2,
        int s1,int *s2,bool iso[],bool *isotail,int m,int big_m, interval imu,
        const interval b[],interval d);
//  Computation of isolation for KS equation:
// ON INPUT:
//  inerr[k] - k=m+1,..,big_m - current bounds
//    for k>big_m - the bound is c1/k^s1
//  b[k] - a bilinear term   for m < k <= big_m
//   and  d/k^{s1 -2} for k > big_m
//   m - is the number of frozen modes
//   iso[k], m<k <= big_m - tells that we have an isolation in k-th directon
//       *isotail - an isolation for all k>big_m
// ON OUTPUT:
//    err[k] - contains new bounds for m <k <=big_m
//      for k > big_m - the bound is given by c2/k^s2
//   iso[k], m<k <= big_m - tells wether we have an isolation in k-th directon
//       *isotail - an isolation for all k>big_m
////////////////
//  we compute the isolation for
//     - l_k  q_k + Left(b[k])  <
//            \frac{d q_k}{dt} < - l_k q_k + Right(b[k])
///  where l_k=imu *k^4 - k^2
//     then:
//       u_k= Right(b[k])/ l_k
//       d_k= Left(b[k])/ l_k
//  u_k and d_k are new isolating bounds, err[k]=interval(d_k,uk_k)
///////////////
//  as for far tail (for k > big_m) it may happen then when we iterate bounds
//   and obtained higer power in the denomnator, the our estimate get worse and
//   we cannot verify the isolation with the present formulas, but we know that
// it is there (by the construction). In this case the value of *isotail -
//       is unchanged and if it was true (on input), then is also true on output


void  ksgalerr_point(const interval q[],const interval& C,const int s,
          const int m,const int big_m,interval err[]);
// AN RIGOROUS FUNCTION
// computes the difference between the galerkin projection and the actual
//  vector field after projection. i.e.
//   Q_k F(p + q) - Q_k F(p)  \subset err[k]
//  where F -  KS vector field,
//   p - is defined by q[1..m]
//   and q[m+1,\dots,big_m] - is the near tail in the self-consistent apriori bounds:
//      |q[k]| < C/k^s for k > big_m - is the far tail



void riggalerrks(interval err[],int galdim,const interval q[],int m,int big_m,interval c,
    int s,const interval boxes[][MAX_DIM+1],int boxnum);
// computes  difference between the galerkin projection and the actual
//  vector field after projection, the projection is <galdim>-dimensional i.e.
//   Q_k F(p + q) - Q_k F(p)  \subset err[k], for k=1..<galdim>
//  where F -  KS vector field,
//  the set, S, over which it is computed is given
//   \sum_{i=1}^{boxnum} boxes[i] \times \Pi q[m < i \leq big_M ] \times_{i > M} [\pm C/i^s]
//  q[k>m],C,s -  the tail  for selfconsistent apriori bounds:
//      q[m+1,...,big_m] and |q[k]| < C/k^s for k > M(=big_m)
////////////
//  galdim - the dimension Galerkin projection
///  p \in P_{galdim}S, q \in (I-P_{galdim})S
////////
/////////////////
  // it must hold : <galdim> <= big_m, dim(boxes[i])=m
  // usually galdim==m


void inittaberr(interval err[],int big_m,interval imu, interval rho,interval rho1,
    interval* c, int* s);
  //initiates bounds for trajectory using JKT approach
  // first from ball in L_2 (rho) and in H_1 (rho1)
  // and then Gronwall type bounds:
  //     4sqrt(2 * pi rho rho1^{3}}/(mu  - k^{-2})


void rigkserr(bool usecurrvaltail,interval mu,const interval boxes[][MAX_DIM+1],
      int boxnum,  int m, int big_m,  bool *tail_isol_ok, interval galerr[],
              interval e[],interval *c,  int *s,int iternum  = 10  );
  // computes bounds for higher modes by iteration of Growall-type bounds
  // the set on which we perform computations is covered by <boxes>[1..<n>]
  //  the number of boxes is equal to <boxnum>
  //  <m> - is the dimension of Galerkin projection and also the dimension
  //      of <boxes[i]>
  //  <big_m>  (the variable M in my papers)-- defining the dimension
  //        in tail computations , must be      big_m>= m, but usually <big_m> is few times <m>,
  //  if (usecurrvaltail) then the computation starts with the values
  //              e,c,s - from input, in this case the input value
  //              of <tail_isol_ok> matters - tells if with current boxes,e,c,s
  //               we have already an isolation for k>m - this could have been
  //               done using some other approach..
  //       the tail is represented as follows:   e[k] - for m < k <= big_m
  //                                        |e[k]| \leq  C/k^s  for  big_m < k
  //  else
  //    otherwise bounds are generated from some guesses based on the norm estimates
  // ---------------------
  // on output the following data are send to a file:
  //   a tail description and the Galerkin projection errors in the vector field
  //   resulting from the tail
  //  tail_isol_ok -  is set to true - when self-consistent apriori bounds
  //               were found
  //    and then galerr[k] - contains the error term of the k-th coordinate
  //      comming from neglected modes via galerkin projection
  //   The new tail is returned in  e,c,s as follows
  //   e[k] - for m < k <= big_m
  //   |e[k]| \leq  C/k^s  for  big_m < k


int uniqueness_lognorm(KuraShi &kurashi,LinCC &coch,int cellnum,
     int cells[][MAX_DIM+1],const interval boxes[][MAX_DIM+1],
    int boxnum,int m,int Big_M,interval e[],interval c,int s,
    DerData *dffxipdata);
//  uniqueness, logarithmic norm and stability/instability computation around fixed point
// the set on which we perform computations is covered by <boxes>[1..<m>]
//  the number of boxes is equal to <boxnum>
  //  coch - a linearlizing coordinates  ; warning: boxes are in standard coordinates
  //      dim(boxes)=dim(coch)= <m>
  //  cellnum - number of blocks in linearization - in dim 1--m
  //       a block is 1-dimensional for a real eigenvalue
  //       and is 2-dimensional for a complex eigenvalue
  //     degenerate situations are not treated here - the number
  //    of eigenvectors matches the number  of eignevalues
  //  cells[i][1]..cells[i][cells[i][0]] - contain the indices of variables
  //       in the block
  //
  //  <m> - is the dimension of Galerkin projection and also the dimension
  //      of <boxes[i]> and it must coincide with kurashi.dim
  //  <Big_M> - M - defining the dimension in tail computations , must be
  //     Big_M>= m, but usually <Big_M> is a few times <m>
  //   the tail is defined by  e,c,s as follows
  //   e[k] - for m < k <= Big_M
  //   |e[k]| \leq  C/k^s  for  Big_M < k
  // -----------------------------------
  //  the function returns 1 - when we have uniqueness of fixed point
  //                       0 - otherwise
  //   on output *dffxipdata - contains both logarithmic norms and uniqueness data
  //                         in the block-decomposition infinity norm
  //      ; IT IS ASSUMED that all blocks are at most 2_D-dimensional
  //      this assumption is explictly used in computations of diagonal terms only




int ks_instability_full_test(KuraShi &kurashi,LinCC &coch,int cellnum,
     int cells[][MAX_DIM+1],const interval boxes[][MAX_DIM+1],
    int boxnum,int m,int Big_M,interval e[],interval c,int s,
    DerData *dffixpdata);
//  an instability computation around fixed point
// the set on which we perform computations is covered by <boxes>[1..<m>]
//  the number of boxes is equal to <boxnum>
  //  coch - a linearlizing coordinates  ; warning: boxes are in standard coordinates
  //      dim(boxes)=dim(coch)= <m>
  //  cellnum - number of blocks in linearization - in dim 1--m
  //       a block is 1-dimensional for a real eigenvalue
  //       and is 2-dimensional for a complex eigenvalue
  //     degenerate situations are not treated here - the number
  //    of eigenvectors matches the number  of eignevalues
  //  cells[i][1]..cells[i][cells[i][0]] - contain the indices of variables
  //       in the block
  //
  //  <m> - is the dimension of Galerkin projection and also the dimension
  //      of <boxes[i]> and it must coincide with kurashi.dim
  //  <Big_M> - M - defining the dimension in tail computations , must be
  //     Big_M>= m, but usually <Big_M> is a few times <m>
  //   the tail is defined by  e,c,s as follows
  //   e[k] - for m < k <= Big_M
  //   |e[k]| \leq  C/k^s  for  Big_M < k
  // -----------------------------------
  //  the function returns 1 - when we have established an instability of fixed point
  //   on output *dffixpdata - contains data


bool ks_cone_condition(KuraShi &kurashi,const LinCC &coch,
    const interval boxes[][MAX_DIM+1], int boxnum,int m,int Big_M,
    interval e[],interval c,int s,  DerData *dffixpdata);
//  a cone condition computation around fixed point
// the set on which we perform computations is covered by <boxes>[1..<m>]
//  the number of boxes is equal to <boxnum>
  //  coch - a linearlizing coordinates  ; warning: boxes are in standard coordinates
  //      dim(boxes)=dim(coch)= <m>
  //
  //  <m> - is the dimension of Galerkin projection and also the dimension
  //      of <boxes[i]> and it must coincide with kurashi.dim
  //  <Big_M> - M - defining the dimension in tail computations , must be
  //     Big_M>= m, but usually <Big_M> is a few times <m>
  //   the tail is defined by  e,c,s as follows
  //   e[k] - for m < k <= Big_M
  //   |e[k]| \leq  C/k^s  for  Big_M < k
  // -----------------------------------
  //  the function returns true - when we have established the cone-conditions
  //  in the neighborhood covered by <box> of fixed point
  //   on output *dffixpdata - contains data


#endif
