#include<iostream>
#include<fstream>
using namespace std;

#include<cstdlib>

#include"defo.h"
#include"zdynsys.h"
#include"multiint.h"
#include"zgfiles.h"
#include"coordc.h"




// constants describing the numerical methods used in the integration of ODE
const int TAYLORM =1 ;
const int  RUNGEKUTTA  =2 ;

char *num_method_name(char name[],int maxlen,int method)
{
  strcpy(name,"");
  if(method==TAYLORM)
    strncpy(name,"Taylor method",maxlen-1);
  if(method==RUNGEKUTTA)
    strncpy(name,"Runge-Kutta method",maxlen-1);
  return name;
}

int whatnumericalmethod()
{
  const int buflen=80;
  char buffer[buflen];
  cout << "1 - " << num_method_name(buffer,buflen,TAYLORM) << endl;
  cout << "2 - " << num_method_name(buffer,buflen,RUNGEKUTTA) << endl;
  int ret;
  cin >> ret;
  #if defined(WIN95) && defined(KRAK)
  cout << endl;
  #endif
  if((ret <1) || (ret>2)){
    ret=TAYLORM;
    cout << "using " << num_method_name(buffer,buflen,TAYLORM) << endl;
  }
  return ret;
}


OneStepParamType::OneStepParamType()
{
  method=TAYLORM;
  order=4;
  orderdf=4;
  doC0intersection=1;
}

OneStepParamType::OneStepParamType(int pmethod,int porder,int porderdf,
           int pdoC0intersection):method(pmethod),order(porder),orderdf(porderdf),
           doC0intersection(pdoC0intersection)
{
/*   method=pmethod;
   order=porder;
   orderdf=porderdf;
   doC0intersection=pdoC0intersection;*/
}

void OneStepParamType::write(ostream &report) const
{
  const int buflen=80;
  char buffer[buflen];
  report << num_method_name(buffer,buflen,method) << endl;
  report << "order=" << order << endl;
  if(doC0intersection){
    report << "using the intersection with the dissipative enclosure "  << endl;
  }
  report  << "C1 - order " << orderdf << endl;
}


Par_Poincare::Par_Poincare()
{
}


void Par_Poincare::write(ostream &report) const
{
  report << "the time step h=" << h << endl;
  report << "tmax=" << tmax << endl;
  onesteppar.write(report);
}



void Ode::setdefaultvalues() // the function assigns the default values
                   // for variables in class ODE, it should be called
                   // in constructors of heirs of this class
{
  dissipative=0;
  lognormused=NORM_MAX;
  enc_order=1;
  divsecreg=4;
}

void Ode::assignvalues(Ode *pode)
    // the function assigns  for variables
                     // in class ODE from pode, the following variables
                     // are assigned values: dissipative, enc_order,lognormused,divsecreg
{
  dissipative=pode->dissipative;
  lognormused=pode->lognormused;
  enc_order=pode->enc_order;
  divsecreg=pode->divsecreg;
}

void Ode::field(const double x[],double f[])
{
  for(int i=1;i<=dim;i++)
    f[i]=vfield(i,x);
}

void Ode::field(const interval x[],interval f[])
{
  for(int i=1;i<=dim;i++)
    f[i]=vfield(i,x);
}


int Ode::adjustenergy(double /*x*/[])
// used to be sure that eventual integrals are conserved (for example - energy
//   for Hamiltonian systems
{
  return 1;
}

int Ode::adjustenergy(interval /*x*/[])
// used to be sure that eventual integrals are conserved (for example - energy
//   for Hamiltonian systems
{
  return 1;
}

int  Ode::adjustenergy(double /* x*/[],double /*sec*/[])
//  the function assumes that <x> is on the section <sec>
//  we adjust conserved quatities so that we stay on the section <sec>
{
  return 1;
}




int Ode::adjustvect(double /*x*/[],double /* tanvect*/[])
//  tangvect is supposed to be in the tangent space at x; when our ode is
//   on manifold
// used for example for Hamiltonian systems
{
  return 1;
}

int Ode::adjustvect(double /*x*/[],double tanvect[],double sec[])
//  tangvect is supposed to be in the tangent space at x; when our ode is
//   on manifold
// used for example for Hamiltonian systems
// if section is nontrivial, then we want <tanvect> to be on section
{
  double tsec[MAX_DIM+1];
  for(int i=1;i<=dim;i++)
    tsec[i]=sec[i];
  if(vect_normalize(tsec,dim,NORM_EUCL)){
    orthogonalcomponent(tanvect,tsec,dim);
  }
  return(1);
}

double Ode::timestep(double h)
{
  return h;
}


interval  Ode::timestep(interval h)
{
  return h;
}


void Ode::pert_fromsegdist(interval encl[],interval dist[],interval h,
        int useperturbation,interval perturbation[])
    // if (useperturbation) then the vector field is given by f(x)+perturbation
    // the function assumes that both a set and its image are contained in
    // <encl> for time [0,h]
    //  then computes distance from segment joing end points and a point
    //  from trajectory
    //  on return dist[i] = 0 if x_i - coordinate is monotonic inside <encl>
    //            and is nonzero otherwise and equal to the above mentioned
    //            distance between segment and entire trajectory
    // useful in section region; perturbation important for differential
    //             inclusions
{
//  interval vectf[MAX_DIM+1];
  interval *vectf;
  vectf=new interval[dim+1];
  for(int i=1; i<=dim; i++)
    vectf[i]=vfield(i,encl);

  if(useperturbation){
    for(int i=1;i<=dim;i++)
      vectf[i]+=perturbation[i];
  }


//  interval Dfi[MAX_DIM+1];
  interval *Dfi;
  Dfi=new interval[dim+1];
  for(int i=1;i<=dim;i++){
    if(vectf[i].contains(0.0)){
      for(int j=1;j<=dim;j++)
        Dfi[j]=dvfield(i,j,encl);
      interval di=0.0;
      for(int j=1;j<=dim;j++)
        di+=Dfi[j]*vectf[j];
      di=abs(di)*power(h,2)/2;
      dist[i]=di;
    } else dist[i]=0.0; // i-th coordinate is monotonic
  }
  delete [] Dfi;
  delete [] vectf;
}

void Ode::fromsegdist(interval encl[],interval dist[],interval h)
     // the function assumes that both a set and its image are contained in
    // <encl> for time [0,h]
    //  then computes distance from segment joing end points and a point
    //  from trajectory
    //  on return dist[i] = 0 if x_i - coordinate is monotonic inside <encl>
    //            and is nonzero otherwise and equal to the above mentioned
    //            distance between segment and entire trajectory
    // useful in section region
{
  pert_fromsegdist(encl,dist,h,0,NULL);
}

void Ode::pert_fromsegdist(interval encl[],interval dist[],interval h,
         interval mac[][MAX_DIM+1],int useperturbation,interval perturbation[])
    // if (useperturbation) then the vector field is given by f(x)+perturbation
   // the function assumes that both a set and its image are contained in
    // <encl> for time [0,h]
    //  then computes distance from segment joining end points and a point
    //  from trajectory, BUT EXPRESSES IT IN NEW COORDINATES m (new coordinates)
    //  - the matrix pcc->intv_m; new coordinates are given by y=m*(x-P)
    //    <encl> - is given on input in old coordinates
    //  on return dist[i] = 0 if x_i - coordinate is monotonic inside <encl>
    //            and is nonzero otherwise and equal to the above mentioned
    //            distance between segment and entire trajectory
    // useful in the section region, when other coordinates are used
    //          and for differential inclusions (useperturbation)
{
//   interval vectf[MAX_DIM+1];
  interval* vectf;
  vectf=new interval[dim+1];
  for(int i=1; i<=dim; i++)
    vectf[i]=vfield(i,encl);
  if(useperturbation){
    for(int i=1;i<=dim;i++)
      vectf[i]+=perturbation[i];
  }

//  interval df[MAX_DIM+1][MAX_DIM+1];
  intervalmatrix1 *df;
  df=new intervalmatrix1[dim+1];
  for(int k=1;k<=dim;k++)
    for(int i=1;i<=dim;i++)
      df[k][i]=dvfield(k,i,encl);

  interval dff[MAX_DIM+1];
  multAV(df,vectf,dff,dim); // dff=df*vectf

  multAV(mac,vectf,vectf,dim); //  vectf - is now expressed in new coordinates
  multAV(mac,dff,dff,dim); //  dff - is now expressed in new coordinates

  for(int i=1;i<=dim;i++){
    if(vectf[i].contains(0.0)){
      dist[i]=abs(dff[i])*power(h,2)/2;
    } else dist[i]=0.0; // i-th coordinate is monotonic
  }
  delete [] df;
  delete [] vectf;
}


void Ode::fromsegdist(interval encl[],interval dist[],interval h,
         interval mac[][MAX_DIM+1])
   // the function assumes that both a set and its image are contained in
    // <encl> for time [0,h]
    //  then computes distance from segment joining end points and a point
    //  from trajectory, BUT EXPRESSES IT IN NEW COORDINATES m (new coordinates)
    //  - the matrix pcc->intv_m; new coordinates are given by y=m*(x-P)
    //    <encl> - is given on input in old coordinates
    //  on return dist[i] = 0 if x_i - coordinate is monotonic inside <encl>
    //            and is nonzero otherwise and equal to the above mentioned
    //            distance between segment and entire trajectory
    // useful in the section region, when other coordinates are used
{
  pert_fromsegdist(encl,dist,h,mac,0,NULL);
}


void Ode::varfromsegdist(interval encl[],interval dist[],
      interval Dencl[][MAX_DIM+1],interval Ddist[][MAX_DIM+1],
      interval h)
   // the function assumes that for a set <x=\varphi(t,x0)>
     // \varphi([0,h],x) \subset <encl>(=[W1])
    //  V(t+[0,h],x0) \subset Dencl (=[W4] in my notes - 'C1-Lohner ...')
    //  then computes distance from segment joing end points and a point
    //  from trajectory
    //  on return dist[i] = 0 if x_i - coordinate is monotonic inside <encl>,
    //            otherwise it is nonzero  and is equal to the above mentioned
    //            distance between segment and entire trajectory
    //  on return Ddist[i][j] = 0 if D_{i,j} - coordinate is monotonic inside
    //            <encl> x <Dencl>
    //            and is nonzero otherwise and equal to the above mentioned
    //            distance between segment and entire trajectory
   // useful in section region
{
  interval vectf[MAX_DIM+1];
  for(int i=1; i<=dim; i++)
    vectf[i]=vfield(i,encl);

  // interval Df[MAX_DIM+1][MAX_DIM+1];
  intervalmatrix1 *Df;
  Df=new intervalmatrix1[dim+1];
  for(int i=1;i<=dim;i++)
    for(int j=1;j<=dim;j++)
      Df[i][j]=dvfield(i,j,encl);

  for(int i=1;i<=dim;i++){
    if(vectf[i].contains(0.0)){
      interval di=0.0;
      for(int j=1;j<=dim;j++)
        di+=Df[i][j]*vectf[j];
      di=abs(di)*power(h,2)/2;
      dist[i]=di;
    } else dist[i]=0.0; // i-th coordinate is monotonic
  }


  //  a computation for variational part
//  interval D2f[MAX_DIM+1][MAX_DIM+1][MAX_DIM+1]; //  D2f[i][j][k] = d^2 f_i/d x_j d x_k (encl)
  intervalmatrix2 *D2f;
  D2f=new intervalmatrix2[dim+1];
  for(int i=1;i<=dim;i++)
    for(int j=1;j<=dim;j++){
      for(int k=1;k<=j;k++){
        D2f[i][j][k]=d2vfield(i,j,k,encl);
        D2f[i][k][j]=D2f[i][j][k];
      }
    }


//  interval Dp[MAX_DIM+1][MAX_DIM+1];  // Dp[i][j]=D'_{ij}
  intervalmatrix1 *Dp;
  Dp=new intervalmatrix1[dim+1];
  ArrArr(Df,Dencl,Dp,dim); // Dp=Df*Dencl
  int sumd2ffdone[MAX_DIM+1][MAX_DIM+1];    // ..[i][k] =1 if a sum for (i,k) was done
  for(int i=1;i<=dim;i++)
    for(int j=1;j<=dim;j++)
      sumd2ffdone[i][j]=0;

//  interval sumd2ff[MAX_DIM+1][MAX_DIM+1];
  intervalmatrix1 *sumd2ff;
  sumd2ff=new intervalmatrix1[dim+1];
  for(int i=1;i<=dim;i++){
    for(int j=1;j<=dim;j++){
      if(Dp[i][j].contains(0.0)){
        interval di=0.0;
        // first part - \Sum_{k,l}\frac{\partial^2 f_i}{\partial x_k x_l}f_l*Dencl_{kj}
        for(int k=1;k<=dim;k++){
          if(!sumd2ffdone[i][k]){
            //  computing the sum \frac{\partial^2 f_{i}}{\partial x_k x_l}*f_l
            interval t=0;
            for(int l=1;l<=dim;l++)
              t+=D2f[i][k][l]*vectf[l];
            sumd2ffdone[i][k]=1;
            sumd2ff[i][k]=t;
          }
        }
        interval t=0.0;
        for(int k=1;k<=dim;k++)
          t+=sumd2ff[i][k]*Dencl[k][j];

        di=t;
        // second part - from partial derivatives with respect to V_{lm}
        //             \frac{\partial f_i}{\partial x_l}*V'_{l,j}
        for(int l=1;l<=dim;l++)
          di+=Df[i][l]*Dp[l][j];

        di=abs(di)*power(h,2)/2;
        Ddist[i][j]=di;
      } else Ddist[i][j]=0.0;  // (i,j) entry in V is monotonic
    }
  }
  delete [] sumd2ff;
  delete [] Dp;
  delete [] D2f;
  delete [] Df;
}


interval Ode::lmfunc(const interval x[])
{
  interval poleVect[MAX_DIM+1];
  for(int i=1; i<=dim; i++)
    poleVect[i]=vfield(i,x);

//  interval Df[MAX_DIM+1][MAX_DIM+1];
  intervalmatrix1 *Df;
  Df=new intervalmatrix1[dim+1];
  for(int i=1; i<=dim; i++)
    for(int j=1; j<=dim; j++)
      Df[i][j]=dvfield(i,j,x);
  interval ret=normMax(poleVect,dim);
  ret=ret*normMax(Df,dim);
  delete [] Df;
  return ret;
}

interval Ode::lognorm(const interval y[])
{
//  interval m[MAX_DIM+1][MAX_DIM+1];
  intervalmatrix1 *m;
  m=new intervalmatrix1[dim+1];
  for(int i=1;i<=dim; i++)
    for(int j=1; j<=dim; j++)
      m[i][j]=dvfield(i,j,y);

  interval retval;
  if(lognormused==NORM_MAX)
    retval=lognormMax(m,dim);

  if(lognormused==NORM_SUM)
    retval=lognormSum(m,dim);
  if(lognormused==NORM_EUCL){
    cerr << "Euclidean logarithmic norm not implemented" << endl;
    #if defined(WIN95) && defined(KRAK)
    waitBt();
    #endif
    exit(EXIT_FAILURE);
  }
  delete [] m;
  return(retval);
}

void  Ode::writeparameters(ostream &report)
{
  report << infostr << endl;
}

double Ode::parameter_value(const char /*par*/[])
  // returns the value of the parameter par
{
  return 0.0;
}

#ifdef KRAK
void  Ode::writeparameters(Frame &frame)
{
  frame << infostr << "\n";
}
#endif

int Ode::dissOneStep(const interval /*x0*/[],interval /*h*/,interval xencl[],interval xstep[],
   int /*use_perturbation */,const interval /* perturb_vect */[])
  //  if use_perturbation then wee add <perturb_vect> to the vector field of Ode
  // on input: \varphi([0,h],x0) \subset xencl
  // On output in  <xstep> - the bounds for \varphi(h,x)_i
  //           obtained from the dissipativity are returned
{
  copyvector(xencl,xstep,dim);
  return 1;
}

int Ode::c1dissOneStep(const interval /*xencl*/[],
                 interval /*h*/, interval Vencl[][MAX_DIM+1],
                  interval Vstep[][MAX_DIM+1])
  // for dissipative ODE's in Vstep returned are  the bounds for
  //         \frac{\partial \varphi_i}{\partial x}_j(h,x0)
  //           obtained from the dissipativity, x0 is the initial condition
  //            for x, and xencl is such that \varphi([0,h],x0) \subset xencl
{
  copymatrix(Vencl,Vstep,dim);
  return 1;
}


int Ode::pert_enclosure(const interval d[],interval h0, interval y[],int useperturbation,
   const interval perturb_vector[],  int nomessages/* =0*/)
// if true is returned then:
//    \varphi([0,h0],d) \subset y
//   \varphi is defined by x' \in field(x) + perturb_vector if useperturb
//   and    \varphi is defined by x' = field(x) if !useperturb
//  h0 - cannot  be  negative if <useperturbation>
 // we want the condition:
   //    y \subset interior(Z)
   //  where: y=d + [0,h]*(vfield(Z) + perturb_vector (if (useperturbation))
  // nomessages=0 -  if the function  fails then a message is send
  //             to a screen and files
  // nomessages=1 -  the failure does not result in message to screen
//  Once we have the enclosure - we stop - there is no refinement

{

 if((Left(h0)<0) && useperturbation){
   cerr << "negative ho - the times step in  'Ode::pert_enclosure' " << endl;
          #if defined(KRAK) && defined(WIN95)
                        waitBt();
     #endif
   exit(EXIT_FAILURE);
 }


 int limit=5;   // tells how many times we iterate the loop of guess and check,
                // inflate the guess and check again

 int retval=0;

 double growth_rate=1.1;
 double init_growth_rate=1.1;

//  interval z[MAX_DIM+1];
 interval *z;
 z=new interval[dim+1];
 interval h=interval(0,1)*h0;


 for(int i=1;i<=dim;i++){
   z[i]=d[i]+h*vfield(i,d);
   if(useperturbation)
     z[i]+=h*perturb_vector[i];
 }

 // we have to be sure that z has nonempty interior
 for(int i=1;i<=dim;i++){
   if(diameter(z[i])==0.0)
     z[i]=inf_increase_interval(z[i]);

 }

 interval s; // an auxliary variable used for center + s  - representation
              // of varoius intervals

 for(int i=1;i<=dim;i++){
   z[i].split(s);  // after this function z[i](old)=z[i](new) + s, but diam(z[i])=0
   z[i]+=init_growth_rate*s;
 }

 int counter=1;
 while(!retval  && (counter <= limit)){
   for(int i=1; i<=dim;i++){
     y[i]=d[i]+h*vfield(i,z);
     if(useperturbation)
       y[i]+=h*perturb_vector[i];
   }
   retval=1;
   for(int i=1;i<=dim; i++){
     if(!in(y[i],z[i],1)){
       retval=0;
       if(counter < limit){ // we update z[i] for a new attempt
                           //  in case of counter==limit
                           //  we want to leave z unchanged for diagnostic reasons
         z[i]=Ihull(z[i],y[i]);
         z[i].split(s); // z[i](old)=z[i](new) + s, but diam(z[i])=0
         z[i]+=s*growth_rate;
       }
     }
   }
   counter++;
 }

 if((!retval) && (!nomessages)){
   cerr << "bad enclosure " << endl;
   diagBadEnclosure(d,h0,y,z,useperturbation,perturb_vector);
 }
 delete [] z;

 return(retval);

}

int Ode::enclosure(const interval d[],interval h0, interval y[],int nomessages/* =0*/)
// if true is returned then:
//    \varphi([0,h0],d) \subset y
//
//  h0 - can be also negative
  // nomessages=0 -  if the function  fails then a message is send
  //             to a screen and files
  // nomessages=1 -  the failure does not result in message to screen
{
  int useperturbation=0;
  return  pert_enclosure(d,h0,y,useperturbation,NULL,nomessages);
}

void Ode::diagBadEnclosure(const interval d[],interval h0,const interval y[],const interval z[],
   int useperturbation,const interval perturb_vect[])
// the function writes to the file <zglerrorlogname>
// the data from failed enlosure procedure
//    this means that it does'nt hold: y=d + [0,h0]vfield(z) \subset interior(z)
///
// it tries to give a reason for this
{

  ofstream report(zglerrorlogname,ios::app);
  entriesinerrorlog=1;
  report << "-------------------------" << endl;
  writeCompInfo(report);
  writedatetime(report);
  writeparameters(report);

  report << "h=" << h0 << endl ;

  report << "d" << endl;
  saveAllCoord(report,d,dim,1);
  report << endl << endl;


  report << "y" << endl;
  saveAllCoord(report,y,dim,1);
  report << endl << endl;


  report << "z" << endl;
  saveAllCoord(report,z,dim,1);
  report << endl << endl;


  interval fd[MAX_DIM+1];
  field(d,fd);
  if(useperturbation){
    for(int i=1;i<=dim;i++)
      fd[i]+=perturb_vect[i];
  }

  report << "fd" << endl;
  saveAllCoord(report,fd,dim,1);
  report << endl;



  // we  detect bad directions
  report << "Bad directions: y[i] is not contained in the interior of z[i] " << endl;
  for(int i=1;i<=dim;i++){
    report << "i=" << i << "   " ;
    if(!in(y[i],z[i],1)) report << "bad  ";
    else                 report << "     ";
    report << "   diam(y[i])=" << diameter(y[i]).right_b() <<
        "   diam(z[i])=" << diameter(z[i]).right_b() ;

    report << endl;
  }
  report << endl << endl;


  int diag=0; // tells whether do additional diagnostics,  and what kind
           // 1 - means hL > 1
           // 2 - 'bad shape"
           // 3 - unknown reason

//  interval df[MAX_DIM+1][MAX_DIM+1];
  intervalmatrix1 *df;
  df=new intervalmatrix1[dim+1];
  for(int i=1;i<=dim;i++)
    for(int j=1;j<=dim;j++)
      df[i][j]=dvfield(i,j,z);


  interval L=normMax(df,dim);

  if (!(abs(h0)*L < 1)){
    diag=1;
    report << "hL > 1" << endl;
  }
  report << "hL=" << (abs(h0*L).right_b()) << endl;

  if(dissipative)
    report << "used dissipative enclosure " << endl;


  if (!diag){
    double diamvect[MAX_DIM+1];
    double dfabs[MAX_DIM+1][MAX_DIM+1];

    for(int i=1;i<=dim; i++)
      diamvect[i]=diameter(z[i]).right_b();

    for(int i=1;i<=dim; i++)
      for(int j=1; j<=dim; j++)
        dfabs[i][j]=abs(df[i][j]).right_b();

    for(int i=1;i<=dim; i++){
      double t=0;
      for(int j=1;j<=dim; j++)
        t+=dfabs[i][j]*diamvect[j];
      t*=fabs(h0);
      if(t > diamvect[i]){
        diag=2;
        report << "h Dfi/Dxj diam(z_j)=" << t << " >  diam(z_i)= " <<
             diamvect[i] <<"  ; i=" << i << endl;
      }
    }

  }
  if(!diag){
    diag=3;
    report << "unknown reason " << endl;
  }
  delete [] df;
  cout << "see the report file for bad enclosure  data - "   <<
    zglerrorlogname  << endl;
}

void Ode::c1enclosure(interval h,const interval xencl[],interval Vencl[][MAX_DIM+1])
 // computation of enclosure for variational part
  //  \frac{\partial \varphi}{\partial x} ([0,h],x)=V([0,h],x)
  //  i.e. starting from identity
  //  this estimate is given  by |Vestm| \leq exp(t*lognorm(df))
  // then we intersect this estimate with Id + [0,h]df(xencl)*Vestm
{
  bool improvevar=true; // tells if we want enclosure for variational part
      // to be improved by one iteration of dv computation

  interval logn;
  if(dim>1)
    logn=lognorm(xencl);
  else logn=Right(dvfield(1,1,xencl));
  interval vestm=exp(Ihull(interval(0),h)*logn); // this is an estimate for
               // an absolute value of V_{ij}
  vestm=Ihull(-vestm,vestm);
  for(int i=1;i<=dim;i++)
    for(int j=1;j<=dim;j++)
      Vencl[i][j]=vestm;

  // an attempt to improve this enclosure  V'=df*V:
  // V \subset Id + h*df*idenclosure
  if(improvevar){
    interval hhull=Ihull(interval(0),h);
//    interval df[MAX_DIM+1][MAX_DIM+1];
    intervalmatrix1 *df;
    df=new intervalmatrix1[dim+1];
    for(int i=1;i<=dim;i++)
      for(int j=1;j<=dim;j++)
          df[i][j]=dvfield(i,j,xencl);

//    interval nencl[MAX_DIM+1][MAX_DIM+1];
    intervalmatrix1 *nencl;
    nencl=new intervalmatrix1[dim+1];

    int maxiterno=dim/2;
    int iter=1;
    while(iter <= maxiterno){
      for(int i=1;i<=dim;i++){
        for(int j=1;j<=dim;j++){
          interval temp=0;
          for(int k=1;k<=dim;k++)
            temp+=df[i][k]*Vencl[k][j];
          nencl[i][j]=hhull*temp;
          if(i==j) nencl[i][j]+=1.0;
        }
      }
      commonpart(Vencl,nencl,Vencl,dim);
      iter++;
    }

    delete [] nencl;
    delete [] df;
  }
}

int Ode::derTPoinc(int dimvar,const double c_sec[],const double x[],const double Dx[][MAX_DIM+1],double gradTPoinc[])
   // on output gradTPoinc[i] = \frac{\partial t_p}{\partial x_i}(x), where
   //   x=P(x_0) for the Poincare map to section such that grad S=c_sec
{
  // if the section is given by \alpha(x)=0, then
  // the formula is: \frac{\partial t_p}{\partial x_i}= - \frac{\grad  \alpha(x) * \frac{\partial \varphi}{\partial x_i}}{\grad \alpha(x) * f(x)}

  int ret=0;
  double *fx;
  fx=new double[dimvar+1];

  for(int i=1;i<=dimvar;i++)
    fx[i]=vfield(i,x);

  double gradAf=scprod(c_sec,fx,dimvar);
  if(gradAf!=0){
    ret=1;
    for(int i=1;i<=dimvar;i++){
      double t=0;
      for(int k=1;k<=dimvar;k++){
        t+=c_sec[k]*Dx[k][i];
      }
      gradTPoinc[i]=-t/gradAf;
    }
  }
  delete [] fx;
  return ret;
}



int   Ode::JacPoincMod(const double c_sec[],const double xn[],double Dx[][MAX_DIM+1],int dimvar)
// the function corrects a jacobian matrix <Dx> of the flow
//  to a Jacobian matrix of Poincare map of <ode> to section <c_sec>
//  <xn> is a point on section <c_sec>, P(..)=xn
//  <c_sec> is the gradient of the section function <alpha>  at <xn>
//  the formula is
//       \frac{P_i}{x_j}= - \frac{f_i}{\grad \alpha * f} *
//     \sum_{k=1}^n \frac{\partial \alpha}{\partial x_k}* \frac{\partial \varphi_k}{\partial x_j}
//       + \frac{\partial \varphi_i}{\partial x_j}
//  here <n> is <dimvar>
//  <alpha> :  \sum c_sec[i]*x[i] + c_sec[0]=0
//     \frac{ \partial \varphi}{\partial x} - Jacobian of the flow - Dx on input

{

  double f[MAX_DIM+1];
  for(int i=1;i<=dimvar;i++)
    f[i]=vfield(i,xn);

  double gradRf=scprod(c_sec,f,dimvar);
  if(gradRf==0) return 0;

  double tt[MAX_DIM+1][MAX_DIM+1];

  for(int i=1; i<=dimvar; i++)
    for(int j=1; j<=dimvar; j++){
      double t=0.0;
      for(int k=1;k<=dimvar; k++)
         t+=c_sec[k]*Dx[k][j];

      t= - f[i] * t/gradRf;  // at this moment t= - f[i]* \frac{\partial t_p}{\partial x_j}

      t+=Dx[i][j];
      tt[i][j]=t;
    }

  copymatrix(tt,Dx,dimvar);  // Dx=tt
  return 1;
}



int   Ode::JacPoincMod(const interval c_sec[],const interval xn[],
           interval Dx[][MAX_DIM+1],int dimVar)
// the function corrects a jacobian matrix <Dx> of the flow
//  to a Jacobian matrix of Poincare map of <ode> to section <c_sec>
//  <xn> is a point on section <c_sec>, P(..)=xn
//  <c_sec> is the gradient of the section function <alpha>  at <xn>
//  the formula is
//       \frac{P_i}{x_j}= - \frac{f_i}{\grad \alpha * f} *
//     \sum_{k=1}^n \frac{\partial \alpha}{\partial x_k}* \frac{\partial \varphi_k}{\partial x_j}
//       + \frac{\partial \varphi_i}{\partial x_j}
//  here <n> is <dimvar>
//  <alpha> :  \sum c_sec[i]*x[i] + c_sec[0]=0
//     \frac{ \partial \varphi}{\partial x} - Jacobian of the flow - Dx on input
///// returns 0 - if {\grad \alpha * f}(xn) contains 0
//            1 - otherwise
{
  int retval=1;
  interval gradRf=0.0;
  for(int i=1;i<= dimVar; i++)
    gradRf+=c_sec[i]*vfield(i,xn);

  if(gradRf.contains(0.0)){
    retval = 0;
    cerr << "JacPoincMod - zero in grad(section)*f " << endl;
    return retval;
  }

//  interval tt[MAX_DIM+1][MAX_DIM+1];
  intervalmatrix1 *tt;
  tt=new intervalmatrix1[dimVar+1];

  for(int i=1; i<=dimVar; i++)
    for(int j=1; j<=dimVar; j++){
      interval t=0.0;
      for(int k=1;k<=dimVar; k++)
         t+=c_sec[k]*Dx[k][j];

      t= - vfield(i,xn) * t/gradRf;  // at this moment t= - f[i]* \frac{\partial t_p}{\partial x_j}

      t+=Dx[i][j];
      tt[i][j]=t;
    }

  for(int i=1; i<=dimVar; i++){
    for(int j=1; j<=dimVar; j++)
      Dx[i][j]=tt[i][j];
  }
  delete [] tt;
  return retval;
}



int Ode::der2TPoinc( int dimvar,const double c_sec[],const double x[],
            const double /*Dx*/ [][MAX_DIM+1], const double DP[][MAX_DIM+1],
            const double gradTPoinc[],const double D2x[][MAX_DIM+1][MAX_DIM+1],
            double d2Tp[][MAX_DIM+1])
// the function computes D2x[i][j]=\frac{\partial t_P}{\partial x_i x_j},
//where P - is the to section map , section alpha(x)=0 (must be!!! of the form c_sec*x+c_sec[0]=0),  x=P(x_0)
//  P is a map from R^n \to R^n
//  Dx=\frac{\partial \varphi }{\partial x},
//  DP=\frac{\partial P }{\partial x},
//  gradTPoinc=\frac{\partial t_P}{\partial x}
//  D2x[i][j][k]=\frac{\partial \varphi_i }{\partial x_j x_k}
//////////////
{
  // we have the following formula
  // <c_sec|f(x)>*\frac{\partial^2 t_p}{\partial x_i x_j} +  <c_sec|df(x)*D_j P>* \frac{t_p}{\partial x_i} + symmetric previous term \\
  // <c_sec|\frac{\partial \varphi}{\partial x_i \partial x_j}> =0
  int ret=0;

  doublematrix1 *dfdp;
  dfdp=new doublematrix1[dimvar+1];

  double *fx;
  fx=new double[dimvar+1];

  doublematrix1* df;
  df=new doublematrix1[dimvar+1];

  for(int i=1;i<=dimvar;i++)
    fx[i]=vfield(i,x);

  for(int i=1;i<=dimvar;i++){
    for(int j=1;j<=dimvar;j++){
      df[i][j]=dvfield(i,j,x);
    }
  }


  double gradAf=scprod(c_sec,fx,dimvar);

  ArrArr(df,DP,dfdp,dimvar); // dfdp=df*DP


  delete [] df;
  df=NULL;
  delete [] fx;
  fx=NULL;

  if(gradAf!=0){
    ret=1;



    for(int i=1;i<=dimvar;i++){
      for(int j=1;j<=dimvar;j++){ // we don't take into the account the symmetry
                 // of second partial derivatives
                // to use it later as the test of the quality of the computations

        // computation of   <c_sec|df(x)*D_j P>* \frac{t_p}{\partial x_i}
        double t1=0;
        for(int k=1;k<=dimvar;k++){
          t1+=c_sec[k]*dfdp[k][j];
        }
        t1=t1*gradTPoinc[i];


         // computation of   <c_sec|df(x)*D_i P>* \frac{t_p}{\partial x_j}
        double t1s=0;
        for(int k=1;k<=dimvar;k++){
          t1s+=c_sec[k]*dfdp[k][i];
        }
        t1s=t1s*gradTPoinc[j];


        // computation of <c_sec|\frac{\partial \varphi}{\partial x_i \partial x_j}>
        double t2;
        t2=0;
        for(int k=1;k<=dimvar;k++){
          t2+=c_sec[k]*D2x[k][i][j];
        }

        d2Tp[i][j]=-(t1+ t1s+t2)/gradAf;

      }
    }

  }
  delete [] dfdp;

  return ret;
}



int Ode::derPoincMod(int dimvar,const double c_sec[],const double x[],const double Dx[][MAX_DIM+1],
  const double D2x[][MAX_DIM+1][MAX_DIM+1],double DP[][MAX_DIM+1],double D2P[][MAX_DIM+1][MAX_DIM+1])
// used to compute derivatives for the Poincare map to a section c_sec*x+c_sec[0]=0
//   we assume that P(x_0)=x; we compute the derivates of P at x
// on input: Dx=\frac{\partial varphi}{\partial x}, D2x[i][j][k]=\frac{\partial varphi_i}{\partial x_j x_k}
// On output:  Dx=\frac{\partial P}{\partial x}, D2P[i][j][k]=\frac{\partial P_i}{\partial x_j x_k}
{
  double *dTp;
  dTp=new double[dimvar+1];
  int ok=derTPoinc(dimvar,c_sec,x,Dx,dTp);
  if(ok){
    double *fx;
    fx=new double[dimvar+1];
    field(x,fx);

    // the formula for DP (is valid for nonafine section too, \grad \alpha (x)=c_sec
    // \frac{\partial P_i}{\partial x_j}=f_i \frac{t_p}{\partial x_j}  +  \frac{\partial \varphi_i}{\partial x_j}

    for(int i=1;i<=dimvar;i++){
      for(int j=1;j<=dimvar;j++){
        DP[i][j]=fx[i]*dTp[j] + Dx[i][j];
      }
    }

    doublematrix1 *d2Tp;
    d2Tp=new doublematrix1[dimvar+1];

    der2TPoinc(dimvar,c_sec,x,Dx,DP,dTp,D2x,d2Tp);




    doublematrix1 *dfdp,*df;
    dfdp=new doublematrix1[dimvar+1];
    df=new doublematrix1[dimvar+1];
    for(int i=1;i<=dimvar;i++)
      for(int j=1;j<=dimvar;j++)
        df[i][j]=dvfield(i,j,x);


    ArrArr(df,DP,dfdp,dimvar);
    delete [] df;
    df=NULL;


    // the formula for the second derivative - the section should affine, because this assumption
    //  is used in the derivation of the \frac{\partial^2 t_p}{\partial x_i x_j}
    //  \frac{\partial P_i}{\partial x_j \partial x_k} =
    // (\sum_s \frac{\partial f_i}{\partial x_2}*\frac{\partial P_s}{\partial x_k})*\frac{\partial t_p}{\partial x_j} + symmetric term (j,k)
    //  + f_i \frac{\partial^2 t_p}{\partial x_j \partial x_k} + \frac{\partial^2 \varphi}{\partial x_j \partial x_k}

    for(int i=1;i<=dimvar;i++){
      for(int j=1;j<=dimvar;j++){
 for(int k=1;k<=dimvar;k++){
          D2P[i][j][k]=dfdp[i][k]*dTp[j] + dfdp[i][j]*dTp[k] + fx[i]*d2Tp[j][k] + D2x[i][j][k];
        }
      }
    }




    delete [] dfdp;
    delete [] d2Tp;
    delete [] fx;
  }

  delete [] dTp;
  return ok;
}


/////////////////////////////////////////////////
/////////////// BackwardOde /////////////////////

BackwardOde::BackwardOde(Ode *ode)
// constructor
{
  setdefaultvalues();
  dim=ode->dim;
  dissipative=0;
  strcpy(infostr,"-");
  strcat(infostr,ode->infostr);
  pode=ode;
  enc_order=ode->enc_order;
  divsecreg=ode->divsecreg;
}

interval BackwardOde::vfield(int k,const interval x[])
 //   vfield(k,x)=f_k(x)
{
  return(-pode->vfield(k,x));
}

double  BackwardOde::vfield(int k,const double x[])
    //   vfield(k,x)=f_k(x)
{
  return(-pode->vfield(k,x));
}

interval  BackwardOde::dvfield(int k, int i,const interval x[])
// dvfield(k,i,x)=\frac{\partial f_k}{\partial x_i}(x)
{
  return(-pode->dvfield(k,i,x));
}

double BackwardOde::dvfield(int k, int i,const double x[])
// dvfield(k,i,x)=\frac{\partial f_k}{\partial x_i}(x)
{
  return(-pode->dvfield(k,i,x));
}

interval BackwardOde::d2vfield(int k, int i, int j,const interval x[])
 // d2vfield(k,i,j,x)=\frac{\partial^2 f_k}{\partial x_i  \partial x_j}(x)
{
  return(-pode->d2vfield(k,i,j,x));
}

double BackwardOde::d2vfield(int k, int i, int j,const double x[])
   // d2vfield(k,i,j,x)=\frac{\partial^2 f_k}{\partial x_i  \partial x_j}(x)
{
  return(-pode->d2vfield(k,i,j,x));
}

int BackwardOde::enclosure(const interval d[],interval h0, interval y[],
               int nomessages /*=0*/)
// if true is returned then:
//    \varphi([0,h0],d) \subset y
//
  // nomessages=0 -  if the function  fails then a message is send
  //             to a screen and files
  // nomessages=1 -  the failure does not result in message to screen
{
  return(pode->enclosure(d,-h0,y,nomessages));
}


/////////////// end of BackwardOde ////////////




void printdf(Ode &ode,double x[],ostream& report,const char* comment)
{
  double f[MAX_DIM+1],df[MAX_DIM+1][MAX_DIM+1];
  for(int k=1;k<=ode.dim;k++)
    f[k]=ode.vfield(k,x);

  for(int k=1;k<=ode.dim;k++)
    for(int i=1;i<=ode.dim;i++)
      df[k][i]=ode.dvfield(k,i,x);


  report << comment << endl;

  report << "dim=" << ode.dim << endl;
  report << "x=(";
  for(int k=1;k<=ode.dim;k++){
    if(k>1) report << ",";
    report << x[k] ;
  }
  report << ")" << endl << endl;

  report << "f(x)=(";
  for(int k=1;k<=ode.dim;k++){
    if(k>1) report << ",";
    report << f[k] ;
  }
  report << ")" << endl << endl;

  report << "linear part - df " << endl;
  for(int k=1;k<=ode.dim;k++){
    for(int i=1;i<=ode.dim;i++)
      report << df[k][i] << "  " ;
    report << endl;
  }


}



void printdf(Ode &ode,double x[],const char* fname,  const char* comment)
{
  ofstream  report(fname);
  printdf(ode,x,report,comment);
}





int changenumberofpoints(int pnum)
{
  int npnum;
  npnum=pnum;
  cout << "the number of points=" << pnum <<
      "   press 1  to change it, 0 - leave unchanged ";
  int c;
  cin >> c;
  #if defined(WIN95) && defined(KRAK)
  cout << endl;
  #endif
  if(c){
    cout << "pnum=" ;
    cin >> npnum ;
    #if defined(WIN95) && defined(KRAK)
    cout << endl;
    #endif
  }
  return(npnum);
}



double whatbias()
{
  double bias;
  cout << "we want to discard the part of trajectory for t<= bias =" ;
  cin >> bias;
  #if defined(WIN95) && defined(KRAK)
  cout << endl;
  #endif
  return bias;
}

double whattmax()
{
  double tmax;
  cout << "tmax = " ;
  cin >> tmax;
  #if defined(WIN95) && defined(KRAK)
  cout << endl;
  #endif
  return tmax;
}



void UpGlobalEstm(estimates *gl,estimates *loc,int firsttime)
{

 gl->rettime=(firsttime)?loc->rettime:Ihull(gl->rettime,loc->rettime);
 gl->seccrossingtime=(firsttime)?
   loc->seccrossingtime:Ihull(gl->seccrossingtime,loc->seccrossingtime);

 gl->sec_timestepnumber=(firsttime)?
   loc->sec_timestepnumber:Ihull(gl->sec_timestepnumber,loc->sec_timestepnumber);

 gl->overestm_crosstime=(firsttime)?
   loc->overestm_crosstime:Ihull(gl->overestm_crosstime,loc->overestm_crosstime);

 gl->sec_ratio=(firsttime)?loc->sec_ratio:
              Ihull(gl->sec_ratio,loc->sec_ratio);

 gl->max_sec_error=(firsttime)?loc->max_sec_error:
               Ihull(gl->max_sec_error,loc->max_sec_error);
 gl->d_max_sec_error=(firsttime)?loc->d_max_sec_error:
               Ihull(gl->d_max_sec_error,loc->d_max_sec_error);

 gl->x_sec_error=(firsttime)?loc->x_sec_error:Ihull(gl->x_sec_error,
                                         loc->x_sec_error);

 gl->sec_LipConst=(firsttime)?loc->sec_LipConst:
             Ihull(gl->sec_LipConst,loc->sec_LipConst);
 gl->sec_IntRkError=(firsttime)?loc->sec_IntRkError:
                            Ihull(gl->sec_IntRkError,loc->sec_IntRkError);

 gl->LipConst=(firsttime)?loc->LipConst:Ihull(gl->LipConst,loc->LipConst);
 gl->IntRkError=(firsttime)?loc->IntRkError:
                            Ihull(gl->IntRkError,loc->IntRkError);
 gl->Rkerror=(firsttime)?loc->Rkerror:Ihull(gl->Rkerror,loc->Rkerror);
 gl->LogN=(firsttime)?loc->LogN:Ihull(gl->LogN,loc->LogN);
 gl->maxNorm=(firsttime)?loc->maxNorm:Ihull(loc->maxNorm,gl->maxNorm);
 gl->LMVal=(firsttime)?loc->LMVal:Ihull(loc->LMVal,gl->LMVal);


 gl->maxgrad=(firsttime)?loc->maxgrad:Ihull(loc->maxgrad,gl->maxgrad);


 int locdim;
 locdim=gl->dim;
 if (locdim < loc->dim)
   locdim=loc->dim;
 for(int i=1; i<=locdim; i++)
   (gl->bounds)[i]=(firsttime)?(loc->bounds)[i]:Ihull((gl->bounds)[i],
                    (gl->bounds)[i]);


 for(int i=2;i<=locdim;i++)
   (gl->ratios_sec_errors)[i]=(firsttime)?(loc->ratios_sec_errors)[i]:
   Ihull((gl->ratios_sec_errors)[i],(gl->ratios_sec_errors)[i]);


 gl->num_of_rep_switches=(firsttime)?loc->num_of_rep_switches:Ihull(gl->num_of_rep_switches,
                                         loc->num_of_rep_switches);
}



void UpdateBounds(interval x[],estimates *estm,int firsttime)
{
 for(int i=1;i<=estm->dim; i++){
   if (firsttime)
     (estm->bounds)[i]=x[i];
   else  (estm->bounds)[i]=Ihull(x[i],(estm->bounds)[i]);
 }
 if (firsttime ){
   estm->maxNorm=norm(x,estm->dim,NORM_EUCL);
   estm->maxgrad=gradNormEucl(x,estm->dim);
 } else
 {
   estm->maxNorm=Ihull(estm->maxNorm,norm(x,estm->dim,NORM_EUCL));
   estm->maxgrad=Ihull(estm->maxgrad,gradNormEucl(x,estm->dim));
 }

}



void UpdateBounds(double x[],estimates *estm,int firsttime)
{
  interval y[MAX_DIM+1];
  for(int i=1;i<=estm->dim; i++)
    y[i]=x[i];
  UpdateBounds(y,estm,firsttime);
}



void UpdateBounds(double x[],estimates *estm,int firsttime,
                   LinCC &coordf)
{
  interval y[MAX_DIM+1];
  for(int i=1;i<=estm->dim; i++)
    y[i]=interval(coordf.coord(x,i));
  UpdateBounds(y,estm,firsttime);
}


void UpdateBounds(interval x[],estimates *estm,int firsttime,
                   LinCC &coordf)
{
  interval y[MAX_DIM+1];
  for(int i=1;i<=estm->dim; i++)
    y[i]=coordf.coord(x,i);
  UpdateBounds(y,estm,firsttime);
}




void initestm(estimates *estm,int edim)
{
  estm->dim=edim;

  estm->maxNorm=0.0;
  int i;

  for(i=1; i<=estm->dim; i++)
    (estm->bounds)[i]=0.0;

  estm->rettime=0.0;
  estm->seccrossingtime=0.0;
  estm->sec_timestepnumber=0.0;
  estm->overestm_crosstime=0.0;

  estm->sec_ratio=0.0;
  estm->max_sec_error=0.0;
  estm->d_max_sec_error=0.0;
  estm->x_sec_error=0.0;
  estm->sec_LipConst=0.0;
  estm->sec_IntRkError=0.0;


  estm->LipConst=0.0;

  estm->IntRkError=0.0;

  estm->Rkerror=0.0;

  estm->LogN=0.0;

  estm->LMVal=0.0;


  estm->maxgrad=0.0;


  for(int i=2;i<=estm->dim;i++)
     estm->ratios_sec_errors[i]=0.0;

  estm->num_of_rep_switches=0;

}



void writeestm(ostream& rhandle,const estimates *estm)
{
  rhandle << "bounds" << endl;
  for(int i=1;i<=estm->dim; i++){
    rhandle << "i=" << i << "   " << (estm->bounds)[i] << "  ";
    double f;

    f=abs((estm->bounds)[i]).right_b();
    rhandle << "fabs()=" << f  << "   " ;
    if (f> 0.0)
      rhandle << " log()=" << log(f);
    rhandle << endl;
  }

  rhandle << "eucl norm estm " << (estm->maxNorm).right_b() << endl;
  rhandle << "grad estm " << (estm->maxgrad).right_b() << endl;

  rhandle.flush();
}


void writeestm(const char *repestm,const estimates *estm)
{
  ofstream rhandle(repestm,ios::out);
  writeestm(rhandle,estm);
  rhandle.close();
}


void  writeGlEstm(ostream& outstr,const estimates *gl,int order,double h)
{
  outstr << "h=" << h << endl;
  outstr << "order=" << order << endl;
  outstr << "Global estimates " << endl;
  outstr << "max norm = " << gl->maxNorm << endl;
  outstr << "grad norm = " << gl->maxgrad << endl;
  outstr << "ret time =" <<gl->rettime << endl;
  outstr << "Lip  =" << gl->LipConst << endl;
  outstr << "ErrDyskret  =" << gl->IntRkError << "*h^" << order << endl;
  outstr << "ErrDyskret  =" << gl->IntRkError*pow(h,order) <<  endl;
  outstr << "One step error  =" << gl->Rkerror << "*h^" << (order+1) << endl;
  #ifdef SWITCH_REP_ON
    outstr << "Number of rep. switches " <<  gl->num_of_rep_switches << endl;
    if(Right(gl->num_of_rep_switches) >0){
      outstr << "In view of rep. switching the division into Lip. part and ErrDyskr - does not make sense" << endl;
    }
  #else
    outstr << "no switching of the representation " << endl;
  #endif

  outstr << "LogNorm = " << gl->LogN << endl;

  outstr << "Section error data: " << endl;
  outstr << "Lip  =" << gl->sec_LipConst << endl;
  outstr << "ErrDyskret  =" << gl->sec_IntRkError << "*h^" << order << endl;
  outstr << "ErrDyskret  =" << gl->sec_IntRkError*pow(h,order) <<  endl;
  outstr << "diameters of return times =" << gl->seccrossingtime << endl;
  outstr << "diam return time/heuristic  (should be close to 1) =" <<
      gl->overestm_crosstime << endl;
  outstr << "the number of time steps intersecting the section: " << gl->sec_timestepnumber
          << endl;

  outstr << "section error ="  << gl->max_sec_error << endl;
  outstr << "Poincare Image/size of set =" << gl->sec_ratio << endl;
  outstr << "added section error ="  << gl->d_max_sec_error << endl;
  outstr << "sec error/heuristic (should be close to 1)" << gl->x_sec_error
       << endl;
/*
  if(gl->dim < 10){
    outstr << "sec error/heuristic (should be close to 1) for each coordinate:" << endl;
    for(int i=2;i<=gl->dim;i++)
      outstr << (gl->ratios_sec_errors[i]).right_b() << "   ";
    outstr << endl;
  }
*/
  outstr << "LM  = " << gl->LMVal << endl;


}
#ifdef KRAK
void  writeGlEstm(FRAME &outstr,const estimates *gl,int order,double h)
{

  outstr << "h=" << h << "\n";
  outstr << "order=" << order << "\n";
  outstr << "Global estimates " << "\n";
  outstr << "max norm = " << gl->maxNorm << "\n";
  outstr << "grad norm = " << gl->maxgrad << "\n";
  outstr << "ret time =" <<gl->rettime << "\n";
  outstr << "Lip  =" << gl->LipConst << "\n";
  outstr << "ErrDyskret  =" << gl->IntRkError << "*h^" << order << "\n";
  outstr << "ErrDyskret  =" << gl->IntRkError*power(h,order) <<  "\n";
  outstr << "One step error  =" << gl->Rkerror << "*h^" << (order+1) << "\n";
  #ifdef SWITCH_REP_ON
  outstr << "Number of rep. switches " <<  gl->num_of_rep_switches << "\n";
  if(Right(gl->num_of_rep_switches) >0){
    outstr << "In view of rep. switching the division into Lip. part and ErrDyskr - does not make sense" << "\n";
  }
 #else
    outstr << "no switching of the representation " << "\n";
  #endif
  outstr << "LogNorm = " << gl->LogN << "\n";

  outstr << "Section error data: " << "\n";
  outstr << "Lip  =" << gl->sec_LipConst << "\n";
  outstr << "ErrDyskret  =" << gl->sec_IntRkError << "*h^" << order << "\n";
  outstr << "ErrDyskret  =" << gl->sec_IntRkError*power(h,order) <<  "\n";
  outstr << "diameters of return times =" << gl->seccrossingtime << "\n";
  outstr << "diam return time/heuristic  (should be close to 1) =" <<
      gl->overestm_crosstime << "\n";
  outstr << "the number of time steps intersecting the section: " << gl->sec_timestepnumber
          << "\n";
  outstr << "section error ="  << gl->max_sec_error << "\n";
  outstr << "Poincare Image/size of set =" << gl->sec_ratio << "\n";
  outstr << "added section error ="  << gl->d_max_sec_error << "\n";
  outstr << "sec error/heuristic (should be close to 1)" << gl->x_sec_error
       << "\n";
/*  if(gl->dim < 10){
    outstr << "sec error/heuristic (should be close to 1) for each coordinate:" << "\n";
    for(int i=2;i<=gl->dim;i++)
      outstr << (gl->ratios_sec_errors[i]).right_b() << "   ";
    outstr << "\n";
  } */
  outstr << "LM  = " << gl->LMVal << "\n";

}
#endif


void  writeGlEstm(const char* reportname,const  estimates *gl,int order,double h)
{
  ofstream report(reportname,ios::out);
  writeGlEstm(report,gl,order,h);
  writeestm(report,gl);
  report.close();
}

interval linIneq_estm(const interval & x,const interval &exp_lambda_h,
            const interval & N, const interval &lambda)
// consider 1D linear equation x' \in \lambda x + (N^-,N+)
// then for h>0 holds
//   x(h) < (x(0) - N^+/(-\lambda))e^{\lambda t} + N^+/(-\lambda)
//   x(h) > (x(0) - N^-/(-\lambda))e^{\lambda t} + N^-/(-\lambda)
// On input:  x - is x(0) - our inital condition
//        exp_lambda_h=e^{h \lambda}
// On output: interval g is returned such that x \subset interior (g)
{
  interval gip,gim,b;
  // gip=g_i^+
  b=Right(N)/(-lambda);
  gip=(Right(x)-b)*exp_lambda_h+ b;


  // gim=g_i^-
  b=Left(N)/(-lambda);
  gim=(Left(x)-b)* exp_lambda_h+ b;


  return Ihull(gim,gip);

}

interval linIneq_estm_no_exp(const interval & x, const interval & N,
        const interval &lambda, const interval h)
 // consider 1D linear equation x' \in \lambda x + (N^-,N+)
// then for h>0 holds
//   x(h) < (x(0) - N^+/(-\lambda))e^{\lambda t} + N^+/(-\lambda)
//   x(h) > (x(0) - N^-/(-\lambda))e^{\lambda t} + N^-/(-\lambda)
// On input:  x - is x(0) - our inital condition
//        exp_lambda_h=e^{h \lambda}
// On output: interval g is returned such that x \subset interior (g)
{
   return  linIneq_estm(x,exp(lambda*h),N,lambda);
}
