#ifndef _zdynsys_h_
#define _zdynsys_h_

extern const int TAYLORM /*=1 */;
extern const int  RUNGEKUTTA  /*=2 */;

class LinCC;

char *num_method_name(char name[],int maxlen,int method);

int whatnumericalmethod();



class OneStepParamType{
  public:
   int method; // RUNGEKUTTA or TAYLORM
   int order;
   int orderdf;
   int doC0intersection;

   OneStepParamType();
   OneStepParamType(int pmethod,int porder,int porderdf,int pdoC0intersection);
   void virtual write(ostream &report) const;

};



class Par_Poincare{
  public:
   OneStepParamType onesteppar;

   Par_Poincare();
   double tmax;
   double h; // the time step

   void virtual write(ostream &report) const;
};






// x'=f(x)
class Ode{
  public:
  char infostr[40];
  int dim;

  void virtual field(const double x[],double f[]);
  void virtual field(const interval x[],interval f[]);
  // we want here both functions: <field> - computes full vector fiels
  //  and <vfield> - just
  // one coordinate - in case when computing each coordinate separately,
  // will require redoing many times the same auxliary computation

  interval virtual  vfield(int k,const interval x[]) = 0;
  double virtual vfield(int k,const double x[]) = 0;
    //   vfield(k,x)=f_k(x)

  interval  virtual dvfield(int k, int i,const interval x[]) = 0;
  double virtual dvfield(int k, int i,const double x[]) = 0;
  // dvfield(k,i,x)=\frac{\partial f_k}{\partial x_i}(x)

  interval virtual   d2vfield(int k, int i, int j,const interval x[]) = 0;
  double virtual  d2vfield(int k, int i, int j,const double x[]) = 0;
   // d2vfield(k,i,j,x)=\frac{\partial^2 f_k}{\partial x_i  \partial x_j}(x)

//  double virtual lognorm(const double x[]);

  interval virtual lognorm(const interval x[]);

  interval virtual lmfunc(const interval x[]);

  void virtual pert_fromsegdist(interval encl[],interval dist[],interval h,
        int useperturbation,interval perturbation[]);
    // if (useperturbation) then the vector field is givem by f(x)+perturbation
    // the function assumes that both a set and its image are contained in
    // <encl> for time [0,h]
    //  then computes distance from segment joing end points and a point
    //  from trajectory
    //  on return dist[i] = 0 if x_i - coordinate is monotonic inside <encl>
    //            and is nonzero otherwise and equal to the above mentioned
    //            distance between segment and entire trajectory
    // useful in section region; perturbation important for differential
    //             inclusions


  void virtual fromsegdist(interval encl[],interval dist[],interval h);
    // the function assumes that both a set and its image are contained in
    // <encl> for time [0,h]
    //  then computes distance from segment joing end points and a point
    //  from trajectory
    //  on return dist[i] = 0 if x_i - coordinate is monotonic inside <encl>
    //            and is nonzero otherwise and equal to the above mentioned
    //            distance between segment and entire trajectory
    // useful in section region

  void virtual varfromsegdist(interval encl[],interval dist[],
       interval Dencl[][MAX_DIM+1],interval Ddist[][MAX_DIM+1],interval h);
   // the function assumes that for a set <x=\varphi(t,x0)>
     // \varphi([0,h],x) \subset <encl>(=[W1])
    //  V(t+[0,h],x0) \subset Dencl (=[W4] in my notes - 'C1-Lohner ...')
    //  then computes distance from the segment joining the end points and a point
    //  from trajectory
    //  on return dist[i] = 0 if x_i - coordinate is monotonic inside <encl>,
    //            otherwise it is nonzero  and is equal to the above mentioned
    //            distance between segment and entire trajectory
    //  on return Ddist[i][j] = 0 if D_{i,j} - coordinate is monotonic inside
    //            <encl> x <Dencl>
    //            and is nonzero otherwise and equal to the above mentioned
    //            distance between segment and entire trajectory
   // useful in section region


   void  virtual fromsegdist(interval encl[],interval dist[],interval h,
                    interval mac[][MAX_DIM+1]);
    //  the function assumes that both a set and its image are contained in
    // <encl> for time [0,h]
    //  then computes distance from segment joining end points and a point
    //  from trajectory, BUT EXPRESSES IT IN NEW COORDINATES m (new coordinates)
    //  - the matrix pcc->intv_m; new coordinates are given by y=m*(x-P)
    //    <encl> - is given on input in old coordinates
    //  on return dist[i] = 0 if x_i - coordinate is monotonic inside <encl>
    //            and is nonzero otherwise and equal to the above mentioned
    //            distance between segment and entire trajectory
    // useful in the section region, when other coordinates are used

  void pert_fromsegdist(interval encl[],interval dist[],interval h,
         interval mac[][MAX_DIM+1],int useperturbation,interval perturbation[]);
    // if (useperturbation) then the vector field is given by f(x)+perturbation
   // the function assumes that both a set and its image are contained in
    // <encl> for time [0,h]
    //  then computes distance from segment joining end points and a point
    //  from trajectory, BUT EXPRESSES IT IN NEW COORDINATES m (new coordinates)
    //  - the matrix pcc->intv_m; new coordinates are given by y=m*(x-P)
    //    <encl> - is given on input in old coordinates
    //  on return dist[i] = 0 if x_i - coordinate is monotonic inside <encl>
    //            and is nonzero otherwise and equal to the above mentioned
    //            distance between segment and entire trajectory
    // useful in the section region, when other coordinates are used
    //          and for differential inclusions (useperturbation)


  int virtual enclosure(const interval d[],interval h0, interval y[],int nomessages=0);
// if true is returned then:
//    \varphi([0,h0],d) \subset y
//  h0 - can be negative
  // nomessages=0 -  if the function  fails then a message is send
  //             to a screen and files
  // nomessages=1 -  the failure does not result in message to screen

  int virtual pert_enclosure(const interval d[],interval h0,  interval y[],int useperturbation,
    const interval perturb_vector[],  int nomessages =0);
// if true is returned then:
//    \varphi([0,h0],d) \subset y
//   \varphi is defined by x' \in field(x) + perturb_vector if useperturb
//   and    \varphi is defined by x' = field(x) if !useperturb
//  h0 - cannot  be  negative if <useperturbation>
 // we want the condition:
   //    y \subset interior(Z)
   //  where: y=d + [0,h]*(vfield(Z) + perturb_vector (if (useperturbation))
  // nomessages=0 -  if the function  fails then a message is send
  //             to a screen and files
  // nomessages=1 -  the failure does not result in message to screen

  int virtual dissOneStep(const interval x0[],interval h,interval xencl[],interval xstep[],
    int use_perturbation,const interval perturb_vect[]);
    //  if use_perturbation then wee add <perturb_vect> to the vector field of Ode
  // on input: \varphi([0,h],x0) \subset xencl
  // On output in  <xstep> - the bounds for \varphi(h,x)_i
  //           obtained from the dissipativity are returned

  int virtual c1dissOneStep(const interval xencl[],interval h,
               interval Vencl[][MAX_DIM+1],interval Vstep[][MAX_DIM+1]);
  // for dissipative ODE's in Vstep returned are  the bounds for
  //         \frac{\partial \varphi_i}{\partial x}_j(h,x0)
  //           obtained from the dissipativity, x0 is the initial condition
  //            for x, and xencl is such that \varphi([0,h],x0) \subset xencl


  void virtual c1enclosure(interval h,const interval xencl[],
                                 interval Vencl[][MAX_DIM+1]);
    // computation of enclosure for variational part
  //  \frac{\partial \varphi}{\partial x} ([0,h],x)=V([0,h],x)
  //  i.e starting from identity
  //  this estimate is given  by |Vestm| \leq exp(t*lognorm(df))
  // then we intersect this estimate with Id + [0,h]df(xencl)*Vestm


  void virtual diagBadEnclosure(const interval d[],interval h0, const interval y[],
        const interval z[], int useperturbation,const interval perturb_vect[]);
// the function writes to the file <zglerrorlogname>
// the data from failed enlosure procedure
//    this means that it does'nt hold: y=d + [0,h0]vfield(z) \subset interior(z)
///


  int virtual derTPoinc(int dimvar,const double c_sec[],const double x[],const double Dx[][MAX_DIM+1],
    double gradTPoinc[]);
   // on output gradTPoinc[i] = \frac{\partial t_p}{\partial x_i}(x), where
   //   x=P(x_0) for the Poincare map to section such that grad S=c_sec


  int virtual der2TPoinc( int dimvar,const double c_sec[],const double x[],
              const double /*Dx*/[][MAX_DIM+1], const double DP[][MAX_DIM+1],
              const double gradTPoinc[],const double D2x[][MAX_DIM+1][MAX_DIM+1],
              double d2Tp[][MAX_DIM+1]);
   // the function computes D2x[i][j]=\frac{\partial t_P}{\partial x_i x_j},
   //where P - is the to section map , section alpha(x)=0 (must be!!! of the form c_sec*x+c_sec[0]=0),  x=P(x_0)
   //  P is a map from R^n \to R^n
   //  Dx=\frac{\partial \varphi }{\partial x},
   //  DP=\frac{\partial P }{\partial x},
   //  gradTPoinc=\frac{\partial t_P}{\partial x}
   //  D2x[i][j][k]=\frac{\partial \varphi_i }{\partial x_j x_k}
   //////////////


int virtual derPoincMod(int dimvar,const double c_sec[],const double x[],const double Dx[][MAX_DIM+1],
       const double D2x[][MAX_DIM+1][MAX_DIM+1],double DP[][MAX_DIM+1],double D2P[][MAX_DIM+1][MAX_DIM+1]);
  // used to compute derivatives for the Poincare map to a section c_sec*x+c_sec[0]=0
  //   we assume that P(x_0)=x; we compute the derivates of P at x
  // on input: Dx=\frac{\partial varphi}{\partial x}, D2x[i][j][k]=\frac{\partial varphi_i}{\partial x_j x_k}
  // On output:  Dx=\frac{\partial P}{\partial x}, D2P[i][j][k]=\frac{\partial P_i}{\partial x_j x_k}


   int virtual  JacPoincMod(const interval c_sec[],const interval xn[], interval Dx[][MAX_DIM+1],
                int dimVar);
   int virtual JacPoincMod(const double c_sec[],const double xn[], double Dx[][MAX_DIM+1],
                int dimVar);
// Reason for the virtual character: when for example we want to perform C2-computation
//  by C1-integration of variational equations then this correction is different,

// the function corects a jacobian matrix <Dx> of the flow
//  to a Jacobian matrix of Poincare map of <ode> to section <c_sec>
//  <xn> is a point on section <c_sec>, P(..)=xn
//  <c_sec> is the gradient of the section function <alpha>  at <xn>
//  the formula is
//       \frac{P_i}{x_j}= - \frac{f_i}{\grad \alpha * f} *
//     \sum_{k=1}^n \frac{\partial \alpha}{\partial x_k}* \frac{\partial \varphi_k}{\partial x_j}
//       + \frac{\partial \varphi_i}{\partial x_j}
//  here <n> is <dimvar>
//  <alpha> :  \sum c_sec[i]*x[i] + c_sec[0]=0
//     \frac{ \partial \varphi}{\partial x} - Jacobian of the flow - Dx on input




  double virtual timestep(double h);  // a proposed dim-dependent time step

  interval virtual timestep(interval h);  // a proposed dim dependent time step

  int virtual adjustenergy(double x[]); // used to be sure that eventual
              // integrals are conserved

  int virtual adjustenergy(interval x[]); // used to be sure that eventual
              // integrals are conserved

  int virtual adjustenergy(double x[],double sec[]);
           //  the function assumes that <x> is on the section <sec>
           //  we adjust conserved quatities so that we stay on the section <sec>


  int virtual adjustvect(double /*x*/[],double /* tanvect*/[]);
  //  tangvect is supposed to be in the tangent space at x; when our ode is
  //   on manifold
  // used for example for Hamiltonian systems

  int virtual adjustvect(double /*x*/[],double tanvect[],double sec[]);
//  the <tangvect> is supposed to be in the tangent space at x; when our ode is
//   on manifold
// used for example for Hamiltonian systems
// if the section is nontrivial, then we want <tanvect> to be on section


  void virtual writeparameters(ostream &report);
  #ifdef KRAK
    void virtual writeparameters(Frame &frame);
  #endif

  double virtual parameter_value(const char par[]);
  // returns the value of the parameter par

  int dissipative;

  int enc_order;    // the rough-enclosure order, currently only order 1 is supported
  double divsecreg;  // in the section region we divide the current time step
                // by this number

  int  lognormused; // used to construct C1enclosures and estimates for differential
                   // inlusions defined by an ODE;
                   // and C0-computations based on logarithmic norms - highly ineffective

  void setdefaultvalues(); // the function assigns the default values
                   // for variables in class ODE, it should be called
                   // in constructors of heirs of this class

  void assignvalues(Ode *pode); // the function assigns  for variables
                     // in class ODE from pode, the following variables
                     // are assigned values: dissipative, enc_order,lognormused,divsecreg

};

//  ode - integrated backwards
class BackwardOde: public Ode{
  public:

  BackwardOde(Ode *ode);

  interval virtual  vfield(int k,const interval x[]);
  double virtual vfield(int k,const double x[]);
    //   vfield(k,x)=f_k(x)

  interval  virtual dvfield(int k, int i,const interval x[]);
  double virtual dvfield(int k, int i,const double x[]);
  // dvfield(k,i,x)=\frac{\partial f_k}{\partial x_i}(x)

  interval virtual   d2vfield(int k, int i, int j,const interval x[]);
  double virtual  d2vfield(int k, int i, int j,const double x[]);
   // d2vfield(k,i,j,x)=\frac{\partial^2 f_k}{\partial x_i  \partial x_j}(x)


  int virtual enclosure(const interval d[],interval h0, interval y[],int nomessages=0);
// if true is returned then:
//    \varphi([0,h0],d) \subset y
//  h0 - can be negative
//
  // nomessages=0 -  if the function  fails then a message is send
  //             to a screen and files
  // nomessages=1 -  the failure does not result in message to screen

  private:
    Ode *pode;
};

void printdf(Ode &ode,double x[],const char* fname,
   const char* comment);

void printdf(Ode &ode,double x[],ostream& report,const char* comment);



// OTHER FUNCTIONS

int changenumberofpoints(int pnum);

double whatbias();

double whattmax();



typedef struct
 {
  int dim;
  interval maxNorm; // max norm of points from trajectory
  interval bounds[MAX_DIM+1]; // lower and upper bounds for  trajectory
  interval rettime;  // return time
  interval seccrossingtime; //  the diameter of the return times
  interval overestm_crosstime; // gives approximatimely how many times is the
       // crossing time overestimated

  interval sec_timestepnumber; // the number of time steps during
                      // which we are crossing the section

  interval max_sec_error;  // maximum section error
  interval d_max_sec_error; // maximum from the obtained section error
         //  and the estimated error from hitting the section under an angle
  interval sec_ratio; //  the quotient of the size image on section/ to the size
                          // of the set
  interval x_sec_error; // gives approximately the ratio of actual section error
            // to the 'dynamic' (heuristic) estimate
  interval ratios_sec_errors[MAX_DIM+1];

  interval sec_LipConst;  // an estimate for Lipschitz constant for Poinc. map -
                 // ignored expansion the vector field direction
  interval sec_IntRkError;  //  an estimate for error from the numerical method
                 // for Poinc. map - ignored expansion the vector field direction

  interval LipConst; // estimate for Lipschitz constant - without ignoring any coordinate
  interval IntRkError; // transported error of the method is equal h^order * IntRkError
  interval Rkerror;   // one step discretization error

  interval LogN;     // local logaritmic norms
  interval LMVal; // the distance from the segment x, xn in the section region

  interval num_of_rep_switches; // counts the number of switches in a reprezentation


  interval maxgrad; // for KS - gradient in L^2 norm,  sqrt((k*a_k)^2)

} estimates;


void UpGlobalEstm(estimates *gl,estimates *loc,int firsttime);


void UpdateBounds(interval x[],estimates *estm,int firsttime);
void UpdateBounds(double x[],estimates *estm,int firsttime);
void UpdateBounds(interval x[],estimates *estm,int firsttime,LinCC &coordf);
void UpdateBounds(double x[],estimates *estm,int firsttime,LinCC &coordf);

void initestm(estimates *estm,int edim);


void writeestm(ostream& rhandle,const estimates *estm);
void writeestm(const char *repestm,const estimates *estm);

void writeGlEstm(ostream& outstr,const  estimates *gl,int order,double h);
void writeGlEstm(ostream& outstr,const estimates *gl,int order,double h);

#ifdef KRAK
void writeGlEstm(FRAME &outstr,const estimates *gl,int order,double h);
#endif

void writeGlEstm(const char* reportname,const  estimates *gl,int order,double h);


interval linIneq_estm(const interval & x,const interval &exp_lambda_h,
            const interval & N, const interval &lambda);
interval linIneq_estm_no_exp(const interval & x, const interval & N,
        const interval &lambda, const interval h);
// consider 1D linear equation x' \in \lambda x + (N^-,N+)
// then for h>0 holds
//   x(h) < (x(0) - N^+/(-\lambda))e^{\lambda t} + N^+/(-\lambda)
//   x(h) > (x(0) - N^-/(-\lambda))e^{\lambda t} + N^-/(-\lambda)
// On input:  x - is x(0) - our inital condition
//        exp_lambda_h=e^{h \lambda}
// On output: interval g is returned such that x \subset interior (g)




#endif
